"""Version information for ansible-navigator collection doc cache.

.. note::

   ``__version_collection_doc_cache__``:

   Indicates the version of the schema of the collection doc cache
   this is checked during initialization, if the version of the cache
   differs from below, the cache will be rebuilt.  This should be
   incremented when the schema changes and need not correspond to the
   application version, although keeping the major in sync is probably
   not a bad idea to minimize the amount of stale docs in the user's cache
"""
__version_collection_doc_cache__ = "1.0"



# cspell:ignore ftype, chksum
"""Catalog collections within the execution environment."""
from __future__ import annotations

import argparse
import hashlib
import json
import multiprocessing
import os
import re
import subprocess
import sys

from collections import Counter
from collections import OrderedDict
from collections.abc import Generator
from datetime import datetime
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

import yaml

from ansible import __version__ as ansible_version
from ansible import plugins
from ansible.utils.plugin_docs import get_docstring
from yaml.error import YAMLError


try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import SafeLoader  # type: ignore

# Import from the source tree whenever possible. When running
# in an execution environment, and therefore not type checking
# import from the key_value_store which was injected in the path.
# The TYPE_CHECKING conditional prevents mypy from attempting the
# import and causing an import error.
try:
    from ansible_navigator.utils.key_value_store import KeyValueStore
except ImportError:
    if not TYPE_CHECKING:
        from key_value_store import KeyValueStore


PROCESSES = (multiprocessing.cpu_count() - 1) or 1


class CollectionCatalog:
    """A collection cataloger."""

    def __init__(self, directories: list[Path]):
        """Initialize the collection cataloger.

        :param directories: A list of directories that may contain collections
        """
        self._directories: list[Path] = directories
        self._collections: OrderedDict[str, dict] = OrderedDict()
        self._errors: list[dict[str, str]] = []
        self._messages: list[str] = []

    def _catalog_plugins(self, collection: dict) -> None:
        """Catalog the plugins within a collection.

        :param collection: Details describing the collection
        """
        file_checksums = {}

        file_manifest_file = collection.get("file_manifest_file", {}).get("name")
        if file_manifest_file:
            file_path = Path(collection["path"], file_manifest_file)
            if file_path.exists():
                with open(file=file_path, encoding="utf-8") as fh:
                    try:
                        loaded = json.load(fh)
                        file_checksums = {v["name"]: v for v in loaded["files"]}
                    except (JSONDecodeError, KeyError) as exc:
                        self._errors.append({"path": str(file_path), "error": str(exc)})

        exempt = ["action", "module_utils", "doc_fragments"]
        plugin_directory = Path(collection["path"], "plugins")
        if plugin_directory.is_dir():
            plugin_dirs = [
                plugin_dir
                for plugin_dir in plugin_directory.iterdir()
                if plugin_dir.is_dir() and plugin_dir.name not in exempt
            ]
            # builtin modules are found in a sibling of the plugin directory
            if collection["known_as"] == "ansible.builtin":
                plugin_dirs.append(Path(collection["path"], "modules"))

            for plugin_dir in plugin_dirs:
                plugin_type = plugin_dir.name
                if plugin_type == "modules":
                    plugin_type = "module"
                files = list(plugin_dir.glob("**/*.py")) + list(plugin_dir.glob("**/*.yml"))
                filenames = (x for x in files)
                self._process_plugin_dir(
                    plugin_type,
                    filenames,
                    file_checksums,
                    collection,
                )

    def _catalog_roles(self, collection: dict[str, Any]) -> None:
        """Catalog the roles within a collection.

        :param collection: Details describing the collection
        """
        # pylint: disable=too-many-locals

        collection_name: str = collection["known_as"]
        collection["roles"] = []
        roles_directory = Path(collection["path"], "roles")
        if not roles_directory.is_dir():
            return

        for role_directory in roles_directory.iterdir():
            if not role_directory.is_dir():
                continue
            role: dict[str, str | dict[str, Any]] = {
                "short_name": role_directory.name,
                "full_name": f"{collection_name}.{role_directory.name}",
            }
            error_cataloging_role = False

            # Argument spec cataloging, it is not required
            argspec_name = "argument_specs.yml"
            argspec_path = role_directory / "meta" / argspec_name
            role["argument_specs"] = {}
            role["argument_specs_path"] = ""
            error = {"path": str(argspec_path)}
            try:
                with argspec_path.open(encoding="utf-8") as fh:
                    role["argument_specs"] = yaml.load(fh, Loader=SafeLoader)["argument_specs"]
                    role["argument_specs_path"] = str(argspec_path)
            except KeyError:
                error["error"] = f"Malformed {argspec_name} for role in {collection_name}."
                self._errors.append(error)
            except FileNotFoundError:
                error["error"] = f"Failed to find {argspec_name} for role in {collection_name}."
                self._errors.append(error)
            except YAMLError:
                error["error"] = f"Failed to load {argspec_name} for role in {collection_name}."
                self._errors.append(error)

            # Defaults cataloging, it is not required
            defaults_name = "main.yml"
            defaults_path = role_directory / "defaults" / defaults_name
            role["defaults"] = {}
            role["defaults_path"] = ""
            error = {"path": str(defaults_path)}
            try:
                with defaults_path.open(encoding="utf-8") as fh:
                    role["defaults"] = yaml.load(fh, Loader=SafeLoader)
                    role["defaults_path"] = str(defaults_path)
            except FileNotFoundError:
                pass
            except YAMLError:
                error["error"] = f"Failed to load {defaults_name} for role in {collection_name}."
                self._errors.append(error)
                error_cataloging_role = True

            # Meta/main.yml cataloging, it is required
            meta_name = "main.yml"
            meta_path = role_directory / "meta" / meta_name
            role["info"] = {}
            role["info_path"] = ""
            error = {"path": str(meta_path)}
            try:
                with meta_path.open(encoding="utf-8") as fh:
                    role["info"] = yaml.load(fh, Loader=SafeLoader)
                    role["info_path"] = str(meta_path)
            except FileNotFoundError:
                error["error"] = f"Failed to find {meta_name} for role in {collection_name}."
                self._errors.append(error)
                error_cataloging_role = True
            except YAMLError:
                error["error"] = f"Failed to load {meta_name} for role in {collection_name}."
                self._errors.append(error)
                error_cataloging_role = True

            # Readme.md cataloging, it is required
            readme_name = "README.md"
            readme_path = role_directory / readme_name
            role["readme"] = ""
            role["readme_path"] = ""
            error = {"path": str(readme_path)}
            try:
                with readme_path.open(encoding="utf-8") as fh:
                    role["readme"] = fh.read()
                    role["readme_path"] = str(readme_path)
            except FileNotFoundError:
                error["error"] = f"Failed to find {readme_name} for role in {collection_name}."
                self._errors.append(error)
                error_cataloging_role = True

            if not error_cataloging_role:
                collection["roles"].append(role)

    @staticmethod
    def _generate_checksum(file_path: Path, relative_path: Path) -> dict:
        """Generate a standard checksum for a file.

        :param file_path: The path to the file to generate a checksum for
        :param relative_path: The relative path within the collection directory structure
        :returns: Details about the file, including the checksum
        """
        sha256_hash = hashlib.sha256()
        with open(file_path, "rb") as fh:
            for byte_block in iter(lambda: fh.read(4096), b""):
                sha256_hash.update(byte_block)
        res = {
            "name": relative_path,
            "ftype": "file",
            "chksum_type": "sha256",
            "chksum_sha256": sha256_hash.hexdigest(),
            "format": 1,
        }
        return res

    def _process_plugin_dir(
        self,
        plugin_type: str,
        filenames: Generator[Path, None, None],
        file_checksums: dict[str, dict],
        collection: dict,
    ) -> None:
        """Process each plugin within one plugin directory.

        :param plugin_type: The type of plugins
        :param filenames: The filenames of the plugins
        :param file_checksums: The checksums for the plugin files
        :param collection: The details of the collection
        """
        for filename in filenames:
            if str(filename.name).startswith("__"):
                continue

            relative_path = Path(filename).relative_to(collection["path"])
            checksum_dict = file_checksums.get(str(relative_path))
            if not checksum_dict:
                checksum_dict = self._generate_checksum(filename, relative_path)
            checksum = checksum_dict[f"chksum_{checksum_dict['chksum_type']}"]
            collection["plugin_checksums"][checksum] = {
                "path": str(relative_path),
                "type": plugin_type,
            }

    def _one_path(self, directory: Path) -> None:
        """Process the contents of an <...>/ansible_collections/ directory.

        :param directory: The path to collections directory to walk and load
        """
        for directory_path in directory.glob("*/*/"):
            manifest_file = directory_path / "MANIFEST.json"
            galaxy_file = directory_path / "galaxy.yml"
            collection = None
            if manifest_file.exists():
                with open(file=manifest_file, encoding="utf-8") as fh:
                    try:
                        collection = json.load(fh)
                        collection["meta_source"] = "MANIFEST.json"
                    except JSONDecodeError:
                        error = {
                            "path": str(manifest_file),
                            "error": "failed to load MANIFEST.json",
                        }
                        self._errors.append(error)
            elif galaxy_file.exists():
                with open(file=galaxy_file, encoding="utf-8") as fh:
                    try:
                        collection = {"collection_info": yaml.load(fh, Loader=SafeLoader)}
                        collection["meta_source"] = "galaxy.yml"
                    except YAMLError:
                        error = {
                            "path": str(galaxy_file),
                            "error": "failed to load galaxy.yml",
                        }
                        self._errors.append(error)
            if collection:
                collection_name = f"{collection['collection_info']['namespace']}"
                collection_name += f".{collection['collection_info']['name']}"
                collection["known_as"] = collection_name
                collection["plugin_checksums"] = {}
                collection["path"] = str(directory_path)

                runtime_file = directory_path / "meta" / "runtime.yml"
                collection["runtime"] = {}
                if runtime_file.exists():
                    with open(file=runtime_file, encoding="utf-8") as fh:
                        try:
                            collection["runtime"] = yaml.load(fh, Loader=SafeLoader)
                        except YAMLError as exc:
                            self._errors.append({"path": str(runtime_file), "error": str(exc)})

                self._collections[collection["path"]] = collection
            else:
                msg = (
                    f"collection path '{directory_path}' is ignored as it does not"
                    " have 'MANIFEST.json' and/or 'galaxy.yml' file(s)."
                )
                self._messages.append(msg)

    def _find_shadows(self) -> None:
        """Determine which collections are hidden by another installation of the same."""
        collection_list = list(self._collections.values())
        counts = Counter([collection["known_as"] for collection in collection_list])
        for idx, (collection_path, o_collection) in reversed(
            list(enumerate(self._collections.items())),
        ):
            self._collections[collection_path]["hidden_by"] = []
            if counts[o_collection["known_as"]] > 1:
                for i_collection in reversed(collection_list[0:idx]):
                    if i_collection["known_as"] == o_collection["known_as"]:
                        self._collections[collection_path]["hidden_by"].insert(
                            0,
                            i_collection["path"],
                        )

    def process_directories(self) -> tuple[dict, list]:
        """Process each parent directory.

        :returns: All collections found and any errors
        """
        for directory in self._directories:
            collection_directory = directory / "ansible_collections"
            if collection_directory.exists():
                self._one_path(collection_directory)
        self.add_pseudo_builtin()
        for _collection_path, collection in self._collections.items():
            self._catalog_plugins(collection)
            self._catalog_roles(collection)
        self._find_shadows()
        return self._collections, self._errors

    def add_pseudo_builtin(self) -> None:
        """Add the pseudo builtin collection."""
        collection: dict[str, str | list | dict] = {}
        collection["known_as"] = "ansible.builtin"
        collection["plugin_checksums"] = {}
        collection["path"] = str(Path(plugins.__file__).parents[1])
        collection["runtime"] = {}
        collection["meta_source"] = "None"
        collection["collection_info"] = {"version": ansible_version}
        self._collections[str(collection["path"])] = collection
        msg = f"Added ansible.builtin from: {collection['path']}"
        self._messages.append(msg)


def worker(pending_queue: multiprocessing.Queue, completed_queue: multiprocessing.Queue) -> None:
    """Extract the documentation from a plugin, place in completed queue.

    :param pending_queue: A queue with plugins to process
    :param completed_queue: The queue in which extracted documentation will be placed
    """
    # pylint: disable=import-outside-toplevel

    # load the fragment_loader _after_ the path is set
    from ansible.plugins.loader import fragment_loader

    while True:
        entry = pending_queue.get()
        if entry is None:
            break
        collection_name, checksum, plugin_path = entry

        try:
            if ansible_version.startswith("2.9"):
                (doc, examples, returndocs, metadata) = get_docstring(
                    filename=str(plugin_path),
                    fragment_loader=fragment_loader,
                )
            else:
                (doc, examples, returndocs, metadata) = get_docstring(
                    filename=str(plugin_path),
                    fragment_loader=fragment_loader,
                    collection_name=collection_name,
                )

        except Exception as exc:  # noqa: BLE001
            err_message = f"{type(exc).__name__} (get_docstring): {exc!s}"
            completed_queue.put(("error", (checksum, plugin_path, err_message)))
            continue

        try:
            q_message = {
                "plugin": {
                    "doc": doc,
                    "examples": examples,
                    "returndocs": returndocs,
                    "metadata": metadata,
                },
                "timestamp": datetime.utcnow().isoformat(),
            }
            completed_queue.put(("plugin", (checksum, json.dumps(q_message, default=str))))
        except JSONDecodeError as exc:
            err_message = f"{type(exc).__name__} (json_decode_doc): {exc!s}"
            completed_queue.put(("error", (checksum, plugin_path, err_message)))


def identify_missing(collections: dict, collection_cache: KeyValueStore) -> tuple[set, list, int]:
    """Identify plugins missing from the cache.

    :param collections: All plugins found across all collections
    :param collection_cache: The key value interface to a sqlite database
    :returns: Handled and plugins missing from the cache, including a count of plugins
    """
    handled = set()
    missing = []
    plugin_count = 0
    for _collection_path, collection in collections.items():
        for checksum, details in collection["plugin_checksums"].items():
            plugin_count += 1
            if checksum not in handled:
                if checksum not in collection_cache:
                    missing.append(
                        (
                            collection["known_as"],
                            checksum,
                            Path(collection["path"], details["path"]),
                        ),
                    )
                handled.add(checksum)
    return handled, missing, plugin_count


def parse_args() -> tuple[argparse.Namespace, list[Path]]:
    # pylint: disable=used-before-assignment
    """Parse the arguments from the command line.

    :returns: The parsed arguments and all directories to search
    """
    parser = argparse.ArgumentParser(description="Catalog collections.")
    parser.add_argument(
        "-d",
        dest="dirs",
        nargs="+",
        help="search within the specified directories",
        default=current_collection_paths,
    )

    parser.add_argument("-a", dest="adjacent", help="prepended to dirs")
    parser.add_argument(
        "-c",
        dest="collection_cache_path",
        help="path to collection cache",
        required=True,
    )
    parsed_args = parser.parse_args()

    adjacent = vars(parsed_args).get("adjacent")
    directories = [adjacent] + parsed_args.dirs if adjacent else parsed_args.dirs

    directories.extend(reversed(sys.path))

    resolved = []
    for directory in directories:
        realpath = Path(directory).resolve()
        if realpath not in resolved:
            resolved.append(realpath)

    return parsed_args, resolved


def retrieve_collections_paths() -> dict:
    """Retrieve the currently set collection paths.

    :returns: Errors or the configured collection directories
    """
    cmd = ["ansible-config", "dump", "|", "grep", "COLLECTIONS_PATHS"]
    proc_out = run_command(cmd)
    if "error" in proc_out:
        return proc_out
    regex = re.compile(r"^(?P<variable>\S+)\((?P<source>.*)\)\s=\s(?P<current>.*)$")
    parsed = regex.match(proc_out["stdout"])
    if parsed:
        try:
            current = yaml.load(parsed.groupdict()["current"], Loader=SafeLoader)
            return {"result": current}
        except (YAMLError, KeyError) as exc:
            return {"error": str(exc)}
    return {"error": f"corrupt current collection path: {proc_out['stdout']}"}


def retrieve_docs(
    collection_cache: KeyValueStore,
    errors: list,
    missing: list,
    stats: dict,
) -> None:
    # pylint: disable=too-many-locals
    """Extract the docs from the plugins.

    :param collection_cache: The key value interface to a sqlite database
    :param errors: Previous errors encountered
    :param missing: Plugins missing from the collection cache
    :param stats: Statistics related to the collection cataloging process
    """
    pending_queue = multiprocessing.Manager().Queue()
    completed_queue = multiprocessing.Manager().Queue()
    processes = []
    for _proc in range(PROCESSES):
        proc = multiprocessing.Process(target=worker, args=(pending_queue, completed_queue))
        processes.append(proc)
        proc.start()

    for entry in missing:
        pending_queue.put(entry)
    for _proc in range(PROCESSES):
        pending_queue.put(None)
    for proc in processes:
        proc.join()

    while not completed_queue.empty():
        message_type, message = completed_queue.get()
        if message_type == "plugin":
            checksum, plugin = message
            collection_cache[checksum] = plugin
            stats["cache_added_success"] += 1
        elif message_type == "error":
            checksum, plugin_path, error = message
            collection_cache[checksum] = json.dumps({"error": error})
            errors.append({"path": str(plugin_path), "error": error})
            stats["cache_added_errors"] += 1


def run_command(cmd: list) -> dict:
    """Run a command using subprocess.

    :param cmd: The command to run, split
    :returns: Errors and the stdout from the command run
    """
    try:
        proc_out = subprocess.run(
            " ".join(cmd),
            capture_output=True,
            check=True,
            text=True,
            shell=True,
        )
        return {"stdout": proc_out.stdout}
    except subprocess.CalledProcessError as exc:
        return {"error": str(exc)}


def main() -> dict:
    # pylint: disable=protected-access
    # pylint: disable=used-before-assignment
    """Run the collection catalog process.

    :returns: The results from the completed collection cataloging process
    """
    stats = {}
    stats["cache_added_success"] = 0
    stats["cache_added_errors"] = 0

    cc_obj = CollectionCatalog(directories=parent_directories)
    collections, errors = cc_obj.process_directories()
    stats["collection_count"] = len(collections)

    collection_cache_path = Path(args.collection_cache_path).resolve().expanduser()
    collection_cache = KeyValueStore(collection_cache_path)

    handled, missing, plugin_count = identify_missing(collections, collection_cache)
    stats["plugin_count"] = plugin_count
    stats["unique plugins"] = len(handled)
    stats["processed"] = len(missing)

    if missing:
        retrieve_docs(collection_cache, errors, missing, stats)

    cached_checksums = collection_cache.keys()
    stats["cache_length"] = len(collection_cache.keys())

    for _collection_path, collection in collections.items():
        for no_doc in set(collection["plugin_checksums"].keys()) - set(cached_checksums):
            del collection["plugin_checksums"][no_doc]

    collection_cache.close()
    return {
        "collections": collections,
        "errors": errors,
        "stats": stats,
        "messages": cc_obj._messages,
    }


if __name__ == "__main__":
    start_time = datetime.now()

    collection_paths = retrieve_collections_paths()
    if "error" in collection_paths:
        sys.exit(collection_paths["error"])
    else:
        current_collection_paths = collection_paths["result"]

    args, parent_directories = parse_args()

    COLLECTION_SCAN_PATHS = ":".join(str(parent) for parent in parent_directories)
    os.environ["ANSIBLE_COLLECTIONS_PATHS"] = COLLECTION_SCAN_PATHS

    result = main()
    result["stats"]["duration"] = (datetime.now() - start_time).total_seconds()
    result["collection_scan_paths"] = COLLECTION_SCAN_PATHS
    print(json.dumps(result, default=str))



"""Scripts that generates documentation from the settings definitions."""
from __future__ import annotations

import logging

from copy import copy
from pathlib import Path
from re import match

import mkdocs_gen_files

from ansible_navigator.configuration_subsystem import Constants as C
from ansible_navigator.configuration_subsystem import NavigatorConfiguration
from ansible_navigator.configuration_subsystem.definitions import SettingsEntry
from ansible_navigator.utils.functions import oxfordcomma
from ansible_navigator.version import __version__


logger = logging.getLogger(__name__)

DOCS_DIR = Path(__file__).parent.resolve()
PROJECT_DIR = DOCS_DIR.parent.parent
TEST_SETTINGS_FIXTURE = (
    PROJECT_DIR
    / "tests"
    / "fixtures"
    / "unit"
    / "configuration_subsystem"
    / "ansible-navigator.yml"
)
DOCS_SETTINGS_SAMPLE = PROJECT_DIR / "docs" / ".generated" / "ansible-navigator.yml"
DOCS_SETTINGS_DUMP = PROJECT_DIR / "docs" / ".generated" / "settings-dump.md"
DOCS_SUBCOMMANDS_OVERVIEW = PROJECT_DIR / "docs" / ".generated" / "subcommands-overview.md"

APP = "ansible-navigator"

PARAM_TABLE_HEADER = [
    "## {}",
    "",
    "",
]
SUBCOMMAND_TABLE_HEADER = [
    "## {}",
    "",
]


def _mk_row(row: tuple) -> list:
    """Convert a row as a markdown definition list entry.

    :param row: The row tuple, with name, description and list of options
    :returns: a list of markdown strings
    """
    result = f'{row[0]} [¶](#{row[0]}){{: class="headerlink" }} {{ #{row[0]} }}\n\n:   {row[1]}\n\n'
    for entry in row[2:]:
        result += f"    {entry}\n\n"
    return [result]


def md_settings_dump() -> str:
    """Generate a table for each subcommand's settings parameters.

    :returns: A list of tables, one each for each subcommand
    """
    lines: list[str] = []
    table = copy(PARAM_TABLE_HEADER)
    table[0] = table[0].format("General parameters")
    table.append("")

    for entry in NavigatorConfiguration.entries:
        if not isinstance(entry.subcommands, list):
            row = _params_row_for_entry(
                entry=entry,
            )
            table.extend(_mk_row(row))
    lines.append("\n".join(table))
    for subcommand in NavigatorConfiguration.subcommands:
        logger.debug("Processing subcommand: %s", subcommand.name)
        entries = [
            entry
            for entry in NavigatorConfiguration.entries
            if isinstance(entry.subcommands, list) and subcommand.name in entry.subcommands
        ]
        logger.debug("  params %s", tuple(entry.name for entry in entries))
        if entries:
            table = copy(PARAM_TABLE_HEADER)
            table[0] = table[0].format(f"\n\n### Subcommand: {subcommand.name}")
            table.append("")
            for entry in entries:
                row = _params_row_for_entry(
                    entry=entry,
                )
                table.extend(_mk_row(row))
            lines.append("\n".join(table))
            # lines.extend(["", "|", ""])
    return "\n".join(lines)


def _params_row_for_entry(entry: SettingsEntry) -> tuple:
    # pylint: disable=too-many-branches
    """Create a row entry for one settings parameter.

    :param entry: The settings entry for which the row will be generated
    :return: A tuple describing the settings parameter
    """
    if entry.cli_parameters is None:
        cli_parameters = "positional"
    else:
        if entry.cli_parameters.short:
            if entry.cli_parameters.long_override:
                long = entry.cli_parameters.long_override
            else:
                long = f"--{entry.name_dashed}"
            cli_parameters = f"``{entry.cli_parameters.short}`` or ``{long}``"
        else:
            cli_parameters = "positional"

    path = entry.settings_file_path("ansible-navigator")
    indent = 4
    yaml_like = ["", "```yaml title='Settings'", ""]
    for idx, path_part in enumerate(path.split(".")):
        yaml_like.append(f"{(2*idx+indent) * ' '}{path_part}:")
    yaml_like.append("```\n")

    if entry.value.schema_default is not C.NOT_SET:
        default = entry.value.schema_default
    elif entry.value.default is not C.NOT_SET:
        default = entry.value.default
    else:
        default = "No default value set"

    choices = oxfordcomma(entry.choices, "or")
    env_var = entry.environment_variable(APP.replace("-", "_"))

    settings = []
    settings.append(f"**Added in version:** {entry.version_added}")
    if choices:
        settings.append(f"**Choices:** {choices}")
    if default is not None:
        settings.append(f"**Default:** {default}")
    if cli_parameters is not None:
        settings.append(f"**CLI:** {cli_parameters}")
    if env_var is not None:
        settings.append(f"**ENV:** {env_var}")

    settings.extend([f"\n{indent * ' '}".join(yaml_like)])

    row = (entry.name_dashed, entry.short_description, *settings)
    return row


def _subcommands_overview() -> str:
    """Generate the subcommand table.

    :returns: A list of available subcommands
    """
    table = SUBCOMMAND_TABLE_HEADER
    table[0] = table[0].format("Available subcommands")
    table.append("")
    for subcommand in NavigatorConfiguration.subcommands:
        table.append(f"\n{subcommand.name}")
        table.append(f"\n:   {subcommand.description}")
        table.append(f"\n    CLI Example: `ansible-navigator {subcommand.name} --help`")
        table.append(f"\n    Colon command: `:{subcommand.name}`")
        table.append(f"\n    Version added: `:{subcommand.version_added}`")
    return "\n".join(table)


def md_settings_sample() -> str:
    """Generate markdown for navigator settings.

    :returns: a markdown string
    """
    sample_settings = ["```yaml title='Settings'", ""]
    settings = TEST_SETTINGS_FIXTURE.read_text().splitlines()
    not_commented = ["---", "ansible-navigator:", "logging:", "level:"]
    for idx, line in enumerate(settings):
        if idx != 2 and match(r"\s{2}\S", line):
            sample_settings.append("#")
        if not any(nc in line for nc in not_commented):
            sample_settings.append("# " + line)
        else:
            sample_settings.append(line)
    sample_settings.append("```")
    result = "\n".join(sample_settings)
    return result


with mkdocs_gen_files.open(DOCS_SETTINGS_SAMPLE, "w") as f:
    f.write(md_settings_sample())

with mkdocs_gen_files.open(DOCS_SETTINGS_DUMP, "w") as f:
    f.write(md_settings_dump())

with mkdocs_gen_files.open(DOCS_SUBCOMMANDS_OVERVIEW, "w") as f:
    f.write(_subcommands_overview())



"""Working handler, instant 1112065, utf-8 max = 112064."""

from __future__ import annotations

from typing import TYPE_CHECKING

from .curses_window import CursesWindow


if TYPE_CHECKING:
    from .field_working import FieldWorking


class FormHandlerWorking(CursesWindow):
    """Handle form button."""

    def __init__(self, screen, ui_config):
        """Initialize the handler for a form working notification.

        :param screen: A curses window
        :param ui_config: The current user interface configuration
        """
        super().__init__(ui_config=ui_config)
        self._screen = screen

    @staticmethod
    def handle(idx, form_fields: list) -> tuple[FieldWorking, int]:
        """Handle the information field, immediate return.

        :param idx: Index to retrieve specific field
        :param form_fields: List of fields from a form
        :returns: Indexed form fields
        """
        return form_fields[idx], 112065



# cspell:ignore envvars
"""Herein lies the base class for all interaction with ansible-runner.

Attributes common to all ansible-runner calls are defined within the base class.
"""
from __future__ import annotations

import logging
import os
import shutil
import sys
import tempfile

from typing import Any

from ansible_runner import Runner


class Base:
    """Base class for ansible-runner calls."""

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        private_data_dir: str | None = None,
        container_engine: str | None = None,
        execution_environment: bool | None = False,
        execution_environment_image: str | None = None,
        navigator_mode: str | None = None,
        container_volume_mounts: list | None = None,
        container_options: list | None = None,
        container_workdir: str | None = None,
        set_environment_variable: dict | None = None,
        pass_environment_variable: list | None = None,
        host_cwd: str | None = None,
        rotate_artifacts: int | None = None,
        timeout: int | None = None,
    ) -> None:
        """Handle the common argument for the ansible-runner interface class.

        :param container_engine: Container engine used to isolate execution
        :param container_options: List of container options to pass to execution engine
        :param execution_environment: Boolean argument that enable execution environment support
        :param execution_environment_image: Container image to use when running an command
        :param navigator_mode: Valid value is either ``stdout`` or ``interactive``. If value is
            set to ``stdout`` passed the ``stdin`` of current running process
            is passed to ``ansible-runner`` which enables receiving command
            line prompts after the executing the command. If value is set to
            ``interactive`` the ``ansible-navigator` will run using
            text user interface (TUI).
        :param container_volume_mounts: List of bind mounts in the form
            ``host_dir:/container_dir:labels``
        :param container_workdir: The working directory within the container
        :param host_cwd: The current local working directory. If value of execution_environment is
            set to True this path will be volume mounted within the execution environment.
        :param set_environment_variable: Dict of user requested environment variables to set
        :param private_data_dir: The directory containing all runner metadata needed to invoke
            the runner module. Output artifacts will also be stored here for
            later consumption.
        :param pass_environment_variable: List of user requested environment variables to pass
        :param rotate_artifacts: Keep at most n artifact directories
        :param timeout: The timeout value in seconds that will be passed to either ``pexpect`` of
            ``subprocess`` invocation (based on ``runner_mode`` selected) while
            executing command. It the timeout is triggered it will force cancel
            the execution.
        """
        self._logger = logging.getLogger(__name__)

        self._set_private_data_directory(private_data_dir)
        self._ce = container_engine
        self._ee = execution_environment
        self._eei = execution_environment_image
        self._navigator_mode = navigator_mode
        self._set_environment_variable: dict[str, Any] = (
            set_environment_variable if isinstance(set_environment_variable, dict) else {}
        )
        self._pass_environment_variable: list[str] = (
            pass_environment_variable if isinstance(pass_environment_variable, list) else []
        )
        self._host_cwd = host_cwd
        self._rotate_artifacts = rotate_artifacts if isinstance(rotate_artifacts, int) else None
        self._timeout = timeout if isinstance(timeout, int) else None
        self.ansible_runner_instance: Runner
        self.cancelled: bool = False
        self.finished: bool = False
        self.status: str | None = None
        self._runner_args: dict = {}

        # when the ce is podman, set the container user to root
        if self._ce == "podman":
            if container_options:
                container_options.append("--user=root")
            else:
                container_options = ["--user=root"]

        if self._ee:
            self._runner_args.update(
                {
                    "container_image": self._eei,
                    "process_isolation_executable": self._ce,
                    "process_isolation": True,
                    "container_volume_mounts": container_volume_mounts,
                    "container_options": container_options,
                    "container_workdir": container_workdir,
                },
            )
        self._runner_args.update(
            {
                "private_data_dir": self._private_data_dir,
                "json_mode": True,
                "quiet": True,
                "cancel_callback": self.runner_cancelled_callback,
                "finished_callback": self.runner_finished_callback,
                "timeout": self._timeout,
            },
        )
        if self._rotate_artifacts is not None:
            self._runner_args["rotate_artifacts"] = self._rotate_artifacts

        self._add_env_vars_to_args()

        if self._host_cwd:
            # ensure the CWD ends with a trailing slash
            host_cwd = os.path.join(self._host_cwd, "")
            self._runner_args["host_cwd"] = host_cwd

        if self._navigator_mode == "stdout":
            self._runner_args.update(
                {
                    "input_fd": sys.stdin,
                    "output_fd": sys.stdout,
                    "error_fd": sys.stderr,
                },
            )

    def __del__(self):
        """Drop the private_data_dir if it is temporary."""
        if (
            self._private_data_dir_is_tmp
            and self._private_data_dir
            and os.path.exists(self._private_data_dir)
        ):
            self._logger.debug(
                "delete temporary ansible-runner private_data_dir at path %s",
                self._private_data_dir,
            )
            shutil.rmtree(self._private_data_dir, ignore_errors=True)

    @staticmethod
    def _generate_tmp_directory():
        """Generate a tmp directory for artifacts.

        :returns: Temporary directory path
        """
        return tempfile.mkdtemp(prefix="ansible-navigator_")

    def _set_private_data_directory(self, provided: str | None) -> None:
        """Set the private data directory to the provided, the username, or a uuid.

        :param provided: Data directory path
        """
        if isinstance(provided, str):
            if os.access(provided, os.W_OK | os.R_OK | os.X_OK):
                private_data_directory = provided
                source = "user provided"
            else:
                self._logger.debug("Provided private data dir `%s` was not user writable", provided)
                private_data_directory = self._generate_tmp_directory()
                source = "user provided, but changed to tmp location due to permissions"
        else:
            private_data_directory = self._generate_tmp_directory()
            source = "not user provided, used tmp location"

        self._private_data_dir_is_tmp = source != "user provided"
        self._private_data_dir = private_data_directory
        self._logger.debug("private data dir %s: %s", source, self._private_data_dir)

    def runner_cancelled_callback(self):
        """Check by runner to see if it should cancel.

        :returns: Boolean cancelled value
        """
        return self.cancelled

    def runner_finished_callback(self, runner: Runner):
        """Call when runner finishes.

        :param runner: A runner instance
        """
        self.status = runner.status
        self.finished = True

    def _add_env_vars_to_args(self):
        """Add environment variables to runner args."""
        self._runner_args["envvars"] = {
            k: v for k, v in os.environ.items() if k.startswith("ANSIBLE_")
        }
        self._runner_args["envvars"].update(self._set_environment_variable)
        for env_var in self._pass_environment_variable:
            value = os.environ.get(env_var)
            if value is None:
                self._logger.warning(
                    "Pass through environment variable `%s` not currently set, discarded",
                    env_var,
                )
            else:
                self._runner_args["envvars"][env_var] = value



"""Utilities related to the merging of dictionaries."""
from __future__ import annotations

from typing import Optional
from typing import Union


class DictMergeError(Exception):
    """Custom exception for a dict merge error."""


Mergeable = Optional[
    Union[
        bool,
        dict,
        list,
        str,
    ]
]


def in_place_list_replace(left: Mergeable, right: Mergeable):
    """Merge right into left in place and returns merged result.

    If left is a primitive, right will be used
    If left key is a list or tuple right will be used.
    If left and right have a common key, they will be dict merged
    If left is missing a key from right, the key from right will be added

    :param left: Dict to be merged into
    :param right: Dict to be merged from
    :raises DictMergeError: If merging cannot be done
    :return: Merged dict
    """
    key = None
    try:
        if left is None or isinstance(left, (str, int, float, bool)):  # noqa: SIM114
            # Border case for first run or if a is a primitive
            left = right
        elif isinstance(left, (tuple, list)):
            left = right
        elif isinstance(left, dict):
            if isinstance(right, dict):
                for key in right:
                    if key in left:
                        left[key] = in_place_list_replace(left[key], right[key])
                    else:
                        left[key] = right[key]
            else:
                msg = f"Cannot merge non-dict '{right}' into dict '{left}'"
                raise DictMergeError(msg)
        else:
            msg = f"Not implemented '{right}' into '{left}'"
            raise DictMergeError(msg)
    except TypeError as exc:
        msg = f"TypeError '{exc}' in key '{key}' when merging '{right}' into '{left}'"
        raise DictMergeError(msg) from exc
    return left



"""Utilities for working with dictionaries using paths."""
from __future__ import annotations

import copy
import operator

from collections.abc import MutableMapping
from enum import Enum
from functools import reduce


class MergeBehaviors(Enum):
    """The merge behaviors."""

    LIST_LIST_EXTEND = "append list to list"
    LIST_LIST_REPLACE = "replace list with list"
    LIST_APPEND = "append to list"
    LIST_REPLACE = "replace list"
    LIST_SORT = "sort resulting list"
    LIST_UNIQUE = "only unique values in resulting list"
    DICT_DICT_UPDATE = "update left dict with right dict"
    DICT_DICT_REPLACE = "replace left dict with right dict"


def get_with_path(content: MutableMapping, path: str):
    """Get a value from a path in a dictionary.

    :param content: The content of the settings file
    :param path: The path to the value
    :return: The value at the path
    """
    return reduce(operator.getitem, path.split("."), content)


def check_path(content: MutableMapping, path: str):
    """Check if a path exists in a dictionary.

    :param content: The content of the settings file
    :param path: The path to the value
    :return: Whether the path exists
    """
    try:
        get_with_path(content, path)
        return True
    except (KeyError, TypeError):
        return False


def delete_with_path(content: MutableMapping, path: str):
    """Delete a value from a path in a dictionary.

    :param content: The content of the settings file
    :param path: The path to the value
    """
    parts = path.split(".")
    del reduce(operator.getitem, parts[:-1], content)[parts[-1]]


def ascendants_from_path(path: str):
    """Get the ascendants of a path.

    :param path: The path to the value
    :return: The ascendants of the path
    """
    parts = path.split(".")
    return [path.rsplit(".", i)[0] for i in range(len(parts))]


def descendants_to_path(path: str):
    """Get the descendants to a path.

    :param path: The path to the value
    :return: The descendants to the path
    """
    parts = path.split(".")
    return [path.rsplit(".", i)[0] for i in reversed(range(len(parts)))]


def remove_and_delete_empty_ascendants(content: MutableMapping, path: str):
    """Remove and delete empty ascendants.

    :param content: The content of the settings file
    :param path: The path to the value
    """
    ascendants = ascendants_from_path(path)
    delete_with_path(content, ascendants.pop(0))
    while ascendants:
        ascendant = ascendants.pop(0)
        branch_value = get_with_path(content, ascendant)
        if branch_value == {}:
            delete_with_path(content, ascendant)
        else:
            break


def place_at_path(
    behaviors: tuple[MergeBehaviors, ...],
    content: dict,
    path: str,
    value: bool | int | list | float | str | list | dict,
) -> dict:
    """Place a value at a path in a dictionary.

    :param behaviors: The merge behaviors
    :param content: The content of the settings file
    :param path: The path to the value
    :param value: The value to place
    :raises ValueError: If something can't be done
    :return: The updated content
    """
    if (
        MergeBehaviors.DICT_DICT_REPLACE in behaviors
        and MergeBehaviors.DICT_DICT_UPDATE in behaviors
    ):
        msg = "Can't use both DICT_DICT_REPLACE and DICT_DICT_UPDATE behaviors"
        raise ValueError(msg)
    if (
        MergeBehaviors.LIST_LIST_EXTEND in behaviors
        and MergeBehaviors.LIST_LIST_REPLACE in behaviors
    ):
        msg = "Can't use both LIST_LIST_EXTEND and LIST_LIST_REPLACE behaviors"
        raise ValueError(msg)

    copied_content = copy.deepcopy(content)
    nested = copied_content
    if path in ("", None):
        if isinstance(value, dict):
            if MergeBehaviors.DICT_DICT_REPLACE in behaviors:
                return value
            if MergeBehaviors.DICT_DICT_UPDATE in behaviors:
                return {**nested, **value}
        msg = "Cannot place non dict at root of dict"
        raise ValueError(msg)

    for part in path.split("."):
        if part == path.rsplit(".", maxsplit=1)[-1]:
            if isinstance(nested.get(part), list):
                if isinstance(value, list):
                    if MergeBehaviors.LIST_LIST_EXTEND in behaviors:
                        nested[part].extend(value)
                    elif MergeBehaviors.LIST_LIST_REPLACE in behaviors:
                        nested[part] = value
                    else:
                        msg = "No behavior specified for LIST_LIST"
                        raise ValueError(msg)
                else:
                    if MergeBehaviors.LIST_APPEND in behaviors:
                        nested[part].append(value)
                    elif MergeBehaviors.LIST_REPLACE in behaviors:
                        nested[part] = value
                        continue
                    else:
                        msg = "No behavior specified for LIST_*"
                        raise ValueError(msg)

                if MergeBehaviors.LIST_UNIQUE in behaviors:
                    nested[part] = list(dict.fromkeys(nested[part]))
                if MergeBehaviors.LIST_SORT in behaviors:
                    nested[part].sort()
                continue

            if isinstance(nested.get(part), dict) and isinstance(value, dict):
                if MergeBehaviors.DICT_DICT_UPDATE in behaviors:
                    nested[part].update(value)
                elif MergeBehaviors.DICT_DICT_REPLACE in behaviors:
                    nested[part] = value
                else:
                    msg = "No behavior specified for DICT_DICT"
                    raise ValueError(msg)
                continue

            nested[part] = value
        elif part not in nested:
            nested[part] = {}
        nested = nested[part]
    return copied_content


def move_to_path(
    behaviors: tuple[MergeBehaviors, ...],
    content: dict,
    new_path: str,
    old_path: str,
) -> dict:
    """Move a value to a path in a dictionary.

    :param behaviors: The merge behaviors
    :param content: The content of the settings file
    :param old_path: The path to the value
    :param new_path: The path to the value
    :return: The updated content
    """
    copied_content = copy.deepcopy(content)
    if new_path == old_path:
        return copied_content

    value = get_with_path(content=copied_content, path=old_path)
    delete_with_path(content=copied_content, path=old_path)
    updated_content = place_at_path(
        content=copied_content,
        path=new_path,
        value=value,
        behaviors=behaviors,
    )
    return updated_content



"""An interface to use a sqlite database as a key-value store."""
from __future__ import annotations

import sqlite3

from collections.abc import ItemsView
from collections.abc import Iterator
from collections.abc import KeysView
from collections.abc import MutableMapping
from collections.abc import ValuesView
from pathlib import Path


class KVSKeysView(KeysView[str]):
    """A glorified KeysView specific to, and returned by, methods in KeyValueStore."""


class KVSItemsView(ItemsView[str, str]):
    """A glorified ItemsView specific to, and returned by, methods in KeyValueStore."""


class KVSValuesView(ValuesView[str]):
    """A glorified ValuesView specific to, and returned by, methods in KeyValueStore."""


class KeyValueStore(MutableMapping[str, str]):
    """An interface to use a sqlite database as a key-value store."""

    def __init__(self, filename: str | Path):
        """Initialize the key-value store.

        :param filename: The full path to the sqlite database file
        """
        self._path = str(filename)
        self.conn = sqlite3.connect(self._path)
        cursor = self.conn.cursor()
        cursor.execute("CREATE TABLE IF NOT EXISTS kv (key text unique, value text)")

    @property
    def path(self) -> str:
        """Provide the filename where the KVS is stored on disk.

        :returns: The path to the key-value store
        """
        return self._path

    def close(self) -> None:
        """Close the connection to the database."""
        self.conn.commit()
        self.conn.close()

    def open_(self) -> sqlite3.Connection:
        """Establish the connection to the database.

        :returns: A connection to the database
        """
        self.conn = sqlite3.connect(self.path)
        return self.conn

    def __len__(self) -> int:
        """Count the number of keys in the key-value store.

        :returns: The number of keys
        """
        cursor = self.conn.cursor()
        rows = cursor.execute("SELECT COUNT(*) FROM kv").fetchone()[0]
        return rows if rows is not None else 0

    def iterkeys(self) -> Iterator[str]:
        """Yield keys from the key-value store one by one.

        :yields: The keys in the key-value store
        """
        cursor = self.conn.cursor()
        for row in cursor.execute("SELECT key FROM kv"):
            yield row[0]

    def itervalues(self) -> Iterator[str]:
        """Yield values from the key-value store one by one.

        :yields: The values in the key-value store
        """
        cursor = self.conn.cursor()
        for row in cursor.execute("SELECT value FROM kv"):
            yield row[0]

    def iteritems(self) -> Iterator[tuple[str, str]]:
        """Yield items from the key-value store one by one.

        :yields: The key-value store as items (key, value)
        """
        cursor = self.conn.cursor()
        for row in cursor.execute("SELECT key, value FROM kv"):
            yield row[0], row[1]

    def keys(self) -> KVSKeysView:
        """Return all keys in the key-value store.

        :returns: All the keys
        """
        return KVSKeysView(self)

    def values(self) -> KVSValuesView:
        """Return all values in the key-value store.

        :returns: All the values
        """
        return KVSValuesView(self)

    def items(self) -> KVSItemsView:
        """Return all items from the key-value store.

        :returns: The key-value store as items (key, value)
        """
        return KVSItemsView(self)

    # mypy complains about this 'str' because it's more specific than the 'object'
    # that dict's Mapping superclass wants. However, it's correct in our case.
    # Our sqlite table expects string keys (column type 'text'). It's an error to
    # pass something else. If we pass something too weird, the sqlite3 lib will
    # throw an error at runtime.
    def __contains__(self, key: str) -> bool:  # type: ignore[override]
        """Check if a given key is in the key-value store.

        This provides dictionary like  `some_key in` support for the key-value store.

        :param key: The key to search for
        :returns: An indication of the provided key's existence in the key-value store
        """
        cursor = self.conn.cursor()
        return cursor.execute("SELECT 1 FROM kv WHERE key = ?", (key,)).fetchone() is not None

    def __getitem__(self, key: str) -> str:
        """Return a value from the key-value store given a key.

        :param key: The key to find
        :raises KeyError: When the key does not exist in the key-value store
        :returns: The value for the key provided
        """
        cursor = self.conn.cursor()
        item = cursor.execute("SELECT value FROM kv WHERE key = ?", (key,)).fetchone()
        if item is None:
            raise KeyError(key)
        return item[0]

    def __setitem__(self, key: str, value: str) -> None:
        """Place a key-value combination in the key-value store.

        :param key: The key of the combination to set
        :param value: The value of the combination to set
        """
        cursor = self.conn.cursor()
        cursor.execute("REPLACE INTO kv (key, value) VALUES (?,?)", (key, value))

    def __delitem__(self, key: str) -> None:
        """Delete a key-value combination in the key-value store.

        :param key: The key of the entry to delete
        :raises KeyError: When the provided key does not exist in the key-value store
        """
        if key not in self:
            raise KeyError(key)
        cursor = self.conn.cursor()
        cursor.execute("DELETE FROM kv WHERE key = ?", (key,))

    def __iter__(self) -> Iterator[str]:
        """Yield values extracted from the key-value store one by one.

        This provides dictionary like `for key in` support for the key-value store.

        :returns: The keys in the key-value store
        """
        return self.iterkeys()

    def __repr__(self) -> str:
        """Represent the key-value store.

        :returns: Representation of the key-value store
        """
        # Loosely based on collections.OrderedDict#__repr__
        if not self:
            return f"{self.__class__.__name__}()"
        return f"{self.__class__.__name__}({list(self.items())})"



"""Step abstractions for actions."""
from __future__ import annotations

from collections import deque
from collections.abc import Sequence
from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from typing import Any
from typing import Callable
from typing import Generic
from typing import TypeVar


class Step:
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    """One step in the flow of things.

    The ``TypedStep`` below should be used for all new actions.
    """

    def __init__(
        self,
        name: str,
        step_type: str,
        value: list[dict[str, str]],
        columns: list[str] | None = None,
        index: int | None = None,
        select_func: Callable[[], Step] | None = None,
        show_func: Callable[[], None] | None = None,
    ) -> None:
        """Initialize the instance of a step.

        :param name: The name of the step
        :param step_type: The type of step
        :param value: The data corresponding to the step
        :param columns: The columns to show when the data is presented as a menu
        :param index: The currently selected entry in the data
        :param select_func: The function to be called when the user selects an entry
        :param show_func: The function to call prior to an entry being shown to the user
        """
        self._index = index
        self._index_changed = False
        self._value = value or []
        self._value_changed = False
        self.columns = columns or []
        self.name = name
        self.next = None
        self.previous = None
        self.select_func = select_func
        self.show_func = show_func
        self.type = step_type

    @property
    def changed(self) -> bool:
        """Return if this has changed.

        :returns: Boolean if index or value has changed
        """
        return self._index_changed or self._value_changed

    @changed.setter
    def changed(self, value: bool) -> None:
        """Set the changed value.

        :param value: The value to set
        :type value: bool
        """
        self._value_check(value, bool)
        self._index_changed = value
        self._value_changed = value

    @property
    def index(self) -> int | None:
        """Return the index.

        :returns: Index (should be ``int``)
        """
        return self._index

    @index.setter
    def index(self, index: int) -> None:
        """Set the index.

        :param index: The index
        :type index: int
        """
        self._value_check(index, (int, type(None)))
        self._index_changed = self._index != index
        self._index = index

    @property
    def selected(self) -> dict[str, Any] | None:
        """Return the selected item.

        :returns: The selected item.
        """
        if self._index is None or not self._value:
            return None
        return self._value[self._index % len(self._value)]

    @property
    def value(self) -> list[dict[str, Any]]:
        """Return the value.

        :returns: The value
        """
        return self._value

    @value.setter
    def value(self, value: list[dict[str, Any]]) -> None:
        """Set the value and changed is needed.

        :param value: List of dicts
        :type value: list
        """
        self._value_check(value, list)
        self._value_changed = self._value != value
        self._value = value

    @staticmethod
    def _value_check(value: Any, want: type | tuple[type, ...]) -> None:
        """Check some expected type against a value's type.

        :param value: Some value for comparison
        :param want: The desired type for value
        :raises ValueError: If value type doesn't match wanted type
        """
        if not isinstance(value, want):
            msg = f"wanted {want}, got {type(value)}"
            raise ValueError(msg)


class StepType(Enum):
    """Type of step, either menu or content."""

    MENU = "menu"
    CONTENT = "content"


T = TypeVar("T")


@dataclass
class TypedStep(Generic[T]):
    # pylint: disable=too-many-instance-attributes
    """One step in the flow of things.

    This should be used for new actions following the pattern of a
    ``TypedDict`` for the user interface as seen in settings.
    """

    name: str
    step_type: StepType
    _index_changed: bool = False
    _index: int | None = None
    _value_changed: bool = False
    _value: Sequence[T] = field(default_factory=list)
    columns: list[str] | None = None
    select_func: Callable[[], TypedStep] | None = None
    show_func: Callable[[], None] | None = None

    @property
    def changed(self) -> bool:
        """Return the changed flag.

        :returns: Indication of change
        """
        return self._index_changed or self._value_changed

    @changed.setter
    def changed(self, value: bool) -> None:
        """Set the changed value.

        :param value: Indication of change
        """
        self._index_changed = value
        self._value_changed = value

    @property
    def index(self) -> int | None:
        """Return the index.

        :returns: The index
        """
        return self._index

    @index.setter
    def index(self, index: int) -> None:
        """Set the index.

        :param index: The index
        """
        self._index_changed = self.index != index
        self._index = index

    @property
    def selected(self) -> T | None:
        """Return the selected item.

        :returns: The selected item
        """
        if self._index is None or not self._value:
            return None
        return self._value[self._index % len(self._value)]

    @property
    def value(self) -> Sequence[T]:
        """Return the value.

        :returns: The value
        """
        return self._value

    @value.setter
    def value(self, value: Sequence[T]) -> None:
        """Set the value and value changed if needed.

        :param value: The value for this instance
        """
        self._value_changed = self._value != value
        self._value = value


class Steps(deque):
    """A custom deque."""

    def back_one(self) -> Any:
        """Pop one step off the dequeue to move backwards.

        :returns: The dequeue after pop() or None if no steps are left
        """
        if self:
            return self.pop()
        return None

    @property
    def current(self) -> Any:
        """Return the current step.

        :returns: The TypedStep (or Step) currently accessed
        """
        return self[-1]

    @property
    def previous(self) -> Any:
        """Return the previous step.

        :returns: The TypedStep (or Step) previously accessed
        """
        return self[-2]



"""Herein lies the ability for ansible-runner to run the ansible-inventory command."""

from __future__ import annotations

from ansible_runner import get_inventory

from .base import Base


class AnsibleInventory(Base):
    # pylint: disable=too-many-arguments
    """Abstraction for ansible-inventory command-line."""

    def fetch_inventory(
        self,
        action: str,
        inventories: list,
        response_format: str | None = None,
        host: str | None = None,
        playbook_dir: str | None = None,
        vault_ids: str | None = None,
        vault_password_file: str | None = None,
    ) -> tuple[str, str]:
        """Run ansible-inventory command and get the inventory related details.

        :param action: Valid values are one of ``graph``, ``host``, ``list``, ``graph`` create
            inventory graph, ``host`` returns specific host info and works as inventory script
            and ``list`` output all hosts info and also works as inventory script.
        :param inventories: List of inventory host paths
        :param response_format: The output format for response. Valid values can be one of ``json``,
            ``yaml``, ``toml``. If ``action`` is ``graph`` only allowed value is ``json``.
        :param host: When ``action`` is set to ``host`` this parameter is used to get the host
            specific information.
        :param playbook_dir: This parameter is used to sets the relative path for the inventory
        :param vault_ids: The vault identity to use
        :param vault_password_file: The vault identity to use
        :returns: A tuple of response and error string (if any)
        """
        return get_inventory(
            action,
            inventories=inventories,
            response_format=response_format,
            host=host,
            playbook_dir=playbook_dir,
            vault_ids=vault_ids,
            vault_password_file=vault_password_file,
            **self._runner_args,
        )



# cspell:ignore fqcn,FQCN

"""This action provides ansible-lint results through ansible-navigator.

Internally, it works by using ansible-runner to execute ansible-lint (optionally
in an execution environment). When running ansible-lint, it passes
``-f codeclimate`` which requests JSON output on stdout. The JSON output contains
a list of issues which we then report in the ansible-navigator UI.

We allow users to dig into an individual issue using the standard :<lineno>
commands, to learn more about the issue.

The full specification for ansible-lint's use of JSON (for the codeclimate
formatter) can be found in src/ansiblelint/formatters/__init__.py in the
ansible-lint codebase.
"""
from __future__ import annotations

import json
import os
import shlex

from collections.abc import Mapping
from datetime import datetime
from enum import IntEnum
from typing import Any

from ansible_navigator.action_base import ActionBase
from ansible_navigator.action_defs import RunStdoutReturn
from ansible_navigator.app_public import AppPublic
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.runner.command import Command
from ansible_navigator.steps import Step
from ansible_navigator.ui_framework import Color
from ansible_navigator.ui_framework import CursesLine
from ansible_navigator.ui_framework import CursesLinePart
from ansible_navigator.ui_framework import CursesLines
from ansible_navigator.ui_framework import Decoration
from ansible_navigator.ui_framework import Interaction
from ansible_navigator.ui_framework import error_notification
from ansible_navigator.ui_framework import nonblocking_notification
from ansible_navigator.ui_framework import success_notification
from ansible_navigator.utils.functions import abs_user_path
from ansible_navigator.utils.functions import remove_ansi
from ansible_navigator.utils.functions import time_stamp_for_file

from . import _actions as actions
from . import run_action


class Severity(IntEnum):
    """A mapping from ansible-lint severity to an integer represented internally.

    Primarily used for sorting.
    """

    INFO = 10
    MINOR = 20
    MAJOR = 30
    CRITICAL = 40
    BLOCKER = 50

    # If ansible-lint ever gives us something we don't expect, return this.
    # In practice, we reverse-sort, so this will always be last.
    UNKNOWN = -1

    @classmethod
    def _missing_(cls, value):
        """Return unknown if ansible-lint ever returns something unexpected.

        :param value: The value
        :returns: A severity unknown
        """
        return Severity.UNKNOWN


def severity_to_color(severity: str) -> int:
    """Convert severity to curses colors.

    :param severity: The severity to convert to a color
    :returns: A color for the severity
    """
    if severity == "minor":
        return Color.MAGENTA
    if severity == "major":
        return Color.YELLOW
    if severity in ("critical", "blocker"):
        return Color.RED
    if severity == "info":
        return Color.CYAN
    return Color.BLACK


def color_menu(colno: int, colname: str, entry: dict[str, Any]) -> tuple[int, int]:
    """Color the menu.

    :param colno: The column number
    :param colname: The column name
    :param entry: The current content entry
    :returns: The foreground and background color
    """
    return (severity_to_color(entry["severity"]), Color.BLACK)


def content_heading(obj: dict, screen_w: int) -> CursesLines:
    """Generate the content heading.

    :param obj: The content for which the heading will be generated
    :param screen_w: The current screen width
    :returns: The content heading
    """
    check_line = f"Message: {obj['check_name']}"
    location = f"Location: {obj['issue_path']}"
    fill_characters = screen_w - len(location) + 1
    location_line = f"{location}{' ' * fill_characters}"

    line_1_part_1 = CursesLinePart(
        column=0,
        string=check_line,
        color=severity_to_color(obj["severity"]),
        decoration=Decoration.NORMAL,
    )
    line_2_part_1 = CursesLinePart(
        column=0,
        string=location_line,
        color=severity_to_color(obj["severity"]),
        decoration=Decoration.UNDERLINE,
    )
    return CursesLines((CursesLine((line_1_part_1,)), CursesLine((line_2_part_1,))))


def filter_content_keys(obj: dict[Any, Any]) -> dict[Any, Any]:
    """Filter out internal keys.

    :param obj: The content from which the content keys will be filtered
    :returns: The content without the internal keys
    """
    ignored_keys = ("fingerprint",)
    return {k: v for k, v in obj.items() if not k.startswith("__") and k not in ignored_keys}


def massage_issue(issue: dict) -> dict:
    """Massage an issue by injecting some useful keys with strings for rendering.

    :param issue: The issue reported
    :returns: The issue reformatted
    """
    massaged = issue.copy()
    massaged["__severity"] = massaged["severity"].capitalize()
    # Version 6.1 and before of ansible-lint used this syntax
    # "check_name": "[fqcn-builtins] Use FQCN for builtin actions."
    # Version 6.2 and later use this syntax
    # check_name": "fqcn-builtins", "description": "Use FQCN for builtin actions."
    if issue["check_name"].startswith("["):
        massaged["__message"] = issue["check_name"].split("]")[1].strip()
    else:
        massaged["__message"] = issue["description"]
    massaged["__path"] = abs_user_path(issue["location"]["path"])
    if isinstance(issue["location"]["lines"]["begin"], Mapping):
        massaged["__line"] = issue["location"]["lines"]["begin"]["line"]
    else:
        massaged["__line"] = issue["location"]["lines"]["begin"]
    massaged["issue_path"] = f"{massaged['__path']}:{massaged['__line']}"
    return massaged


MENU_COLUMNS = [
    "__severity",
    "__message",
    "__path",
    "__line",
]


@actions.register
class Action(ActionBase):
    """Run the lint subcommand."""

    KEGEX = r"^lint(\s(?P<params>.*))?$"

    def __init__(self, args: ApplicationConfiguration):
        """Initialize the action.

        :param args: The current application configuration
        """
        self._modification_times_last_updated: datetime
        self._success_notification_shown: bool = False
        self._issues_menu = Step(
            name="issues_menu",
            step_type="menu",
            columns=MENU_COLUMNS,
            value=[],
            select_func=self._build_issue_content,
        )
        self._current_issue_count = 0
        super().__init__(args=args, logger_name=__name__, name="lint")

    @property
    def is_interactive(self):
        """Determine if interactive.

        :returns: An indication is running in interactive mode
        """
        return self._args.mode == "interactive"

    def _fatal(self, msg: str) -> None:
        """Show a notification if a fatal error has occurred.

        :param msg: The message to display
        :raises RuntimeError: A runtime error if not mode interactive
        """
        self._logger.error(msg)

        if self.is_interactive:
            notification = error_notification(messages=[msg])
            self._interaction.ui.show_form(notification)
        else:
            raise RuntimeError(msg)

    def _run_runner(self) -> tuple[str, str, int]:
        """Spin up runner to run ansible-lint, either in an exec env or not.

        :returns: The output, errors and return code
        """
        kwargs = {
            "container_engine": self._args.container_engine,
            "execution_environment_image": self._args.execution_environment_image,
            "execution_environment": self._args.execution_environment,
            "navigator_mode": self._args.mode,
            "pass_environment_variable": self._args.pass_environment_variable,
            "set_environment_variable": self._args.set_environment_variable,
            "private_data_dir": self._args.ansible_runner_artifact_dir,
            "rotate_artifacts": self._args.ansible_runner_rotate_artifacts_count,
            "timeout": self._args.ansible_runner_timeout,
            "host_cwd": os.getcwd(),
        }

        if isinstance(self._args.execution_environment_volume_mounts, list):
            kwargs["container_volume_mounts"] = self._args.execution_environment_volume_mounts

        if isinstance(self._args.container_options, list):
            kwargs["container_options"] = self._args.container_options

        cmd_args = ["-qq", "--offline"]

        if self.is_interactive:
            cmd_args += [
                "--nocolor",
                "-f",
                "codeclimate",
            ]
        elif not self._args.display_color:
            cmd_args.append("--nocolor")

        if isinstance(self._args.lint_config, str):
            cmd_args += [
                "-c",
                self._args.lint_config,
            ]

        if isinstance(self._args.lintables, str):
            # Does it actually exist?
            if not os.path.exists(self._args.lintables):
                self._fatal(f"The given path `{self._args.lintables}` does not exist.")
                return "", "", 1
            cmd_args.append(self._args.lintables)

        kwargs["cmdline"] = cmd_args

        runner = Command(executable_cmd="ansible-lint", **kwargs)
        return runner.run()

    def run_stdout(self) -> RunStdoutReturn:
        """Execute the ``doc`` request for mode stdout.

        :returns: A message and return code
        """
        self._logger.debug("lint requested in stdout mode")
        _output, _error, return_code = self._run_runner()
        return RunStdoutReturn(message="", return_code=return_code)

    @staticmethod
    def _pull_out_json_or_fatal(stdout: str) -> str | None:
        """
        Attempt to pull out JSON line from ansible-lint raw output.

        Note that stdout and stderr get munged together by docker/podman, so we
        need to do some trickery to try to figure out the actual JSON line vs,
        say, ansible warnings.

        :param stdout: The stdout from the lint invocation
        :returns: The json string
        """
        # We want the last (non empty) line of output. This should hopefully be
        # the JSON we need.
        for line in reversed(stdout.splitlines()):
            if not line:
                continue
            return line
        return None

    def run(self, interaction: Interaction, app: AppPublic) -> Interaction | None:
        """Execute the ``lint`` request for mode interactive.

        :param interaction: The interaction from the user
        :param app: The app instance
        :returns: The pending :class:`~ansible_navigator.ui_framework.ui.Interaction` or
            :data:`None`
        """
        self._logger.debug("lint requested")
        self._prepare_to_run(app, interaction)

        updated = self._update_args(
            ["lint"] + shlex.split(interaction.action.match.groupdict()["params"] or ""),
        )

        if not updated:
            self._prepare_to_exit(interaction)
            return None

        notification = nonblocking_notification(messages=["Linting, this may take a minute..."])
        interaction.ui.show_form(notification)

        self._build_issues_menu()
        if self._current_issue_count == 0:
            self._prepare_to_exit(interaction)
            return None

        self.steps.append(self._issues_menu)

        while True:
            self.update()

            if self._current_issue_count == 0:
                break

            self._interaction.ui.update_status(
                status=f"Issues: {self._current_issue_count}",
                status_color=self._max_severity_color,
            )

            self._take_step()

            if not self.steps:
                break

            if self.steps.current.name == "quit":
                return self.steps.current

        self._prepare_to_exit(interaction)
        return None

    def update(self):
        """Request calling app update, and update modification time if needed."""
        self._calling_app.update()

        # Do this only every 2 seconds
        if (datetime.now() - self._modification_times_last_updated).total_seconds() > 2:
            rerun_lint = self._rerun_needed()
            if rerun_lint:
                self._build_issues_menu()

    @property
    def _max_severity_color(self):
        """Determine the color of the maximum severity issue.

        :returns: The color
        """
        max_severity = max(Severity[i["severity"].upper()].value for i in self._issues_menu.value)
        return severity_to_color(Severity(max_severity).name.lower())

    def _take_step(self) -> None:
        """Take a step based on the current step or step back."""
        result = None
        if isinstance(self.steps.current, Interaction):
            result = run_action(self.steps.current.name, self.app, self.steps.current)
        elif isinstance(self.steps.current, Step):
            if self.steps.current.show_func:
                current_index = self.steps.current.index
                self.steps.current.show_func()
                self.steps.current.index = current_index

            if self.steps.current.type == "menu":
                result = self._interaction.ui.show(
                    obj=self.steps.current.value,
                    columns=self.steps.current.columns,
                    color_menu_item=color_menu,
                )
            elif self.steps.current.type == "content":
                result = self._interaction.ui.show(
                    obj=self.steps.current.value,
                    index=self.steps.current.index,
                    content_heading=content_heading,
                    filter_content_keys=filter_content_keys,
                )

        if result is None:
            self.steps.back_one()
        else:
            self.steps.append(result)

    def _build_issues_menu(self):
        """Build the menu of all issues.

        :returns: Indication of success
        """
        output, _error, return_code = self._run_runner()
        self._logger.debug("Output from ansible-lint run (rc=%d): %s", return_code, output)

        # ansible-lint failed
        if return_code != 0 and "ansible-lint: No such file or directory" in output:
            installed_or_ee = (
                "in the execution environment you are using"
                if self._args.execution_environment
                else "installed"
            )
            self._fatal(
                "ansible-lint executable could not be found. Ensure 'ansible-lint' "
                f"is {installed_or_ee} and try again.",
            )
            self._current_issue_count = 0
            return

        # Could not extract json from response
        out_without_warnings = self._pull_out_json_or_fatal(output)
        if out_without_warnings is None:
            self._current_issue_count = 0
            return

        # De-serialization of json failed
        try:
            raw_issues = json.loads(out_without_warnings)
        except json.JSONDecodeError as exc:
            self._logger.debug("Failed to parse 'ansible-lint' JSON response: %s", str(exc))
            messages = ["Could not parse 'ansible-lint' output."]
            without_ansi = remove_ansi(output)
            messages.extend(without_ansi.splitlines())
            notification = error_notification(messages)
            self._interaction.ui.show_form(notification)
            self._current_issue_count = 0
            return

        # No issues were found
        if len(raw_issues) == 0:
            notification = success_notification(messages=["Congratulations, no lint issues found!"])
            self._interaction.ui.show_form(notification)

        # Update menu
        issues = [massage_issue(issue) for issue in raw_issues]
        self._current_issue_count = len(issues)
        self._issues_menu.value = sorted(
            issues,
            key=lambda i: Severity[i["severity"].upper()],
            reverse=True,
        )

        # Update new content timestamps
        _rerun_lint = self._rerun_needed()
        return

    def _build_issue_content(self):
        """Build the content for one plugin.

        :returns: The plugin's content
        """
        return Step(
            name="issue_content",
            step_type="content",
            value=self.steps.current.value,
            index=self.steps.current.index,
        )

    def _rerun_needed(self) -> bool:
        """Add a modification timestamp to each issue, check for changes.

        :returns: Indication if lint should be rerun
        """
        rerun_lint = False
        checked_paths = []
        for outer_issue in self._issues_menu.value:
            outer_path = outer_issue["__path"]
            if outer_path in checked_paths:
                continue
            checked_paths.append(outer_path)
            unix_ts, iso_ts = time_stamp_for_file(outer_path, self._args.time_zone)
            # Set all issues for the same file
            for inner_issue in self._issues_menu.value:
                inner_path = inner_issue["__path"]
                if inner_path != outer_path:
                    continue

                previous_ts = inner_issue.get("__last_modified")
                inner_issue["last_modified"] = iso_ts
                inner_issue["__last_modified"] = unix_ts

                # The file may be gone or non existent
                if unix_ts is None:
                    continue

                # The may be a first run
                if previous_ts is None:
                    rerun_lint = True
                    continue

                # Has the file changed
                if unix_ts > previous_ts:
                    rerun_lint = True

        self._modification_times_last_updated = datetime.now()
        if rerun_lint:
            self._logger.debug("Files modified")
        return rerun_lint



"""Conditional imports related to python versions."""

# pylint: disable=unused-import

import sys

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    # https://mypy.readthedocs.io/en/stable/common_issues.html#variables-vs-type-aliases
    if sys.version_info >= (3, 10):
        from typing import TypeAlias
    else:
        from typing_extensions import TypeAlias


if sys.version_info < (3, 10):
    import importlib_metadata
else:
    import importlib.metadata as importlib_metadata



"""Definitions for the command runner."""
from __future__ import annotations

import multiprocessing
import subprocess

from dataclasses import dataclass
from dataclasses import field
from queue import Queue
from typing import Callable

from ansible_navigator.utils.definitions import LogMessage


PROCESSES = (multiprocessing.cpu_count() - 1) or 1


@dataclass(frozen=False)
class Command:
    """Data structure for details of a command to be run.

    A ``Command`` is updated after instantiated with details from either
    ``stdout`` or ``stderr``.
    """

    # pylint: disable=too-many-instance-attributes
    identity: str
    command: str
    post_process: Callable
    return_code: int = 0
    stdout: str = ""
    stderr: str = ""
    details: list = field(default_factory=list)
    errors: str = ""
    messages: list[LogMessage] = field(default_factory=list)

    @property
    def stderr_lines(self):
        """Produce a list of stderr lines.

        :returns: A list of stderr lines
        """
        return self.stderr.splitlines()

    @property
    def stdout_lines(self):
        """Produce a list of stdout lines.

        :returns: A list of stdout lines
        """
        return self.stdout.splitlines()


def run_command(command: Command) -> None:
    """Run a command.

    :param command: Command to be run
    """
    try:
        proc_out = subprocess.run(
            command.command,
            capture_output=True,
            check=True,
            text=True,
            shell=True,
        )
        command.return_code = proc_out.returncode
        command.stdout = proc_out.stdout
    except subprocess.CalledProcessError as exc:
        command.return_code = exc.returncode
        command.stdout = str(exc.stdout)
        command.stderr = str(exc.stderr)


def worker(pending_queue: multiprocessing.Queue, completed_queue: multiprocessing.Queue) -> None:
    """Read pending, run, post process, and place in completed.

    :param pending_queue: All pending commands
    :param completed_queue: All completed commands
    """
    while True:
        command = pending_queue.get()
        if command is None:
            break
        run_command(command)
        command.post_process(command)
        completed_queue.put(command)


class CommandRunner:
    """Functionality for running commands."""

    def __init__(self):
        """Initialize the command runner."""
        self._completed_queue: Queue | None = None
        self._pending_queue: Queue | None = None

    @staticmethod
    def run_single_process(commands: list[Command]):
        """Run commands with a single process.

        :param commands: All commands to be run
        :returns: The results from running all commands
        """
        results: list[Command] = []
        for command in commands:
            run_command(command)
            command.post_process(command)
            results.append(command)
        return results

    def run_multi_process(self, commands: list[Command]) -> list[Command]:
        """Run commands with multiple processes.

        Workers are started to read from pending queue.
        Exit when the number of results is equal to the number
        of commands needing to be run.

        :param commands: All commands to be run
        :returns: The results from running all commands
        """
        if self._completed_queue is None:
            self._completed_queue = multiprocessing.Manager().Queue()
        if self._pending_queue is None:
            self._pending_queue = multiprocessing.Manager().Queue()

        self.start_workers(commands)
        results: list[Command] = []
        while len(results) != len(commands):
            results.append(self._completed_queue.get())
        return results

    def start_workers(self, jobs):
        """Start the workers.

        :param jobs: List of commands to be run
        """
        worker_count = min(len(jobs), PROCESSES)
        processes = []
        for _proc in range(worker_count):
            proc = multiprocessing.Process(
                target=worker,
                args=(self._pending_queue, self._completed_queue),
            )
            processes.append(proc)
            proc.start()
        for job in jobs:
            self._pending_queue.put(job)
        for _proc in range(worker_count):
            self._pending_queue.put(None)
        for proc in processes:
            proc.join()



from __future__ import annotations

import json
import os

from typing import Any
from typing import NamedTuple
from typing import TypeVar

from .compiler import Compiler
from .fchainmap import FChainMap
from .reg import _Reg
from .reg import make_reg
from .rules import Rule
from .rules import _Rule
from .utils import uniquely_constructed


T = TypeVar("T")


@uniquely_constructed
class Grammar(NamedTuple):
    scope_name: str
    repository: FChainMap[str, _Rule]
    patterns: tuple[_Rule, ...]

    @classmethod
    def make(cls, data: dict[str, Any]) -> Grammar:
        scope_name = data["scopeName"]
        if "repository" in data:
            # this looks odd, but it's so we can have a self-referential
            # immutable-after-construction chain map
            repository_dct: dict[str, _Rule] = {}
            repository = FChainMap(repository_dct)
            for k, dct in data["repository"].items():
                repository_dct[k] = Rule.make(dct, repository)
        else:
            repository = FChainMap()
        patterns = tuple(Rule.make(d, repository) for d in data["patterns"])
        return cls(scope_name=scope_name, repository=repository, patterns=patterns)


class Grammars:
    def __init__(self, *directories: str) -> None:
        """Initialize an instance of Grammars.

        :param directories: A tuple of strings, each a directory in which grammar files can be found
        """
        self._scope_to_files = {
            os.path.splitext(filename)[0]: os.path.join(directory, filename)
            for directory in directories
            if os.path.exists(directory)
            for filename in sorted(os.listdir(directory))
            if filename.endswith(".json")
        }

        unknown_grammar = {"scopeName": "source.unknown", "patterns": []}
        self._raw = {"source.unknown": unknown_grammar}
        self._file_types: list[tuple[frozenset[str], str]] = []
        self._first_line: list[tuple[_Reg, str]] = []
        self._parsed: dict[str, Grammar] = {}
        self._compiled: dict[str, Compiler] = {}

    def _raw_for_scope(self, scope: str) -> dict[str, Any]:
        try:
            return self._raw[scope]
        except KeyError:
            pass

        grammar_path = self._scope_to_files.pop(scope)
        with open(grammar_path, encoding="UTF-8") as f:
            ret = self._raw[scope] = json.load(f)

        file_types = frozenset(ret.get("fileTypes", ()))
        first_line = make_reg(ret.get("firstLineMatch", "$impossible^"))

        self._file_types.append((file_types, scope))
        self._first_line.append((first_line, scope))

        return ret

    def grammar_for_scope(self, scope: str) -> Grammar:
        try:
            return self._parsed[scope]
        except KeyError:
            pass

        raw = self._raw_for_scope(scope)
        ret = self._parsed[scope] = Grammar.make(raw)
        return ret

    def compiler_for_scope(self, scope: str) -> Compiler:
        try:
            return self._compiled[scope]
        except KeyError:
            pass

        grammar = self.grammar_for_scope(scope)
        ret = self._compiled[scope] = Compiler(grammar, self)
        return ret

    def blank_compiler(self) -> Compiler:
        return self.compiler_for_scope("source.unknown")

    def compiler_for_file(self, filename: str, first_line: str) -> Compiler:
        # didn't find it in the fast path, need to read all the json
        for k in tuple(self._scope_to_files):
            self._raw_for_scope(k)

        _, _, ext = os.path.basename(filename).rpartition(".")
        for extensions, scope in self._file_types:
            if ext in extensions:
                return self.compiler_for_scope(scope)

        for reg, scope in self._first_line:
            if reg.match(first_line, 0, first_line=True, boundary=True):
                return self.compiler_for_scope(scope)

        return self.compiler_for_scope("source.unknown")



"""Implementation of the asynchronous invocation of ``ansible-runner``.

Herein lies the ability to invoke ``ansible-runner`` in an async fashion.
A queue is provided and ``ansible-runner`` uses ``pexpect`` to parse
standard out and error from the command run and populates the
queue with messages.
"""

from copy import deepcopy
from queue import Queue

from ansible_runner import run_command_async

from .command_base import CommandBase


class CommandAsync(CommandBase):
    """A wrapper for the asynchronous runner."""

    def __init__(self, executable_cmd: str, queue: Queue, write_job_events: bool, **kwargs):
        """Initialize the arguments for the ``run_command_async`` interface of ``ansible-runner``.

        For common arguments refer to the documentation of the ``CommandBase`` class.

        :param executable_cmd: The command to be invoked
        :param queue: The queue to post events from ``ansible-runner``
        :param write_job_events: Allows job_events to be processed by ``ansible-runner``
        :param kwargs: The arguments for the async runner call
        """
        self._queue = queue
        self._write_job_events = write_job_events
        super().__init__(executable_cmd, **kwargs)

    def _event_handler(self, event):
        """Handle the event from ansible-runner.

        :param event: The event from ansible-runner
        :returns: The value of ``self._write_job_events``, a boolean
        """
        self._logger.debug("ansible-runner event handle: %s", event)
        new_event = deepcopy(event)
        self._queue.put(new_event)
        return self._write_job_events

    def run(self):
        """Initiate the execution of the runner command in async mode.

        :returns: The runner thread
        """
        self.generate_run_command_args()
        self._runner_args.update({"event_handler": self._event_handler})
        thread, self.ansible_runner_instance = run_command_async(**self._runner_args)
        self.status = "running"
        return thread



"""Logging initialization."""
from __future__ import annotations

import datetime
import logging
import os
import zoneinfo

from .configuration_subsystem import Constants
from .configuration_subsystem.definitions import ApplicationConfiguration


logger = logging.getLogger("ansible_navigator")


class Formatter(logging.Formatter):
    """Format a logging timestamp using a time zone."""

    def __init__(self, *args, **kwargs):
        """Initialize the logging formatter.

        :param args: The arguments
        :param kwargs: The keyword arguments
        """
        self._time_zone = kwargs.pop("time_zone")
        super().__init__(*args, **kwargs)

    def formatTime(self, record: logging.LogRecord, _datefmt: str | None = None) -> str:
        """Format the log timestamp.

        :param record: The log record
        :param _datefmt: The optional date format
        :returns: The timestamp
        """
        if self._time_zone == "local":
            return (
                datetime.datetime.fromtimestamp(record.created, datetime.timezone.utc)
                .astimezone()
                .isoformat()
            )
        return datetime.datetime.fromtimestamp(
            record.created,
            tz=zoneinfo.ZoneInfo(self._time_zone),
        ).isoformat()


def setup_logger(args: ApplicationConfiguration) -> None:
    """Set up the logger.

    :param args: The CLI args
    """
    if os.path.exists(args.log_file) and args.log_append is False:
        os.remove(args.log_file)
    handler = logging.FileHandler(args.log_file)

    time_zone = args.entry("time_zone").value.current
    # When the configuration is rolled back, the time_zone will be C.NOT_SET
    if isinstance(args.time_zone, Constants):
        time_zone = args.entry("time_zone").value.default

    formatter = Formatter(
        fmt="%(asctime)s %(levelname)s '%(name)s.%(funcName)s' %(message)s",
        time_zone=time_zone,
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    log_level = getattr(logging, args.log_level.upper())
    logger.setLevel(log_level)
    logger.info("New %s instance, logging initialized", args.application_name_dashed)

    # set ansible-runner logs
    runner_logger = logging.getLogger("ansible-runner")
    runner_logger.setLevel(log_level)
    runner_logger.addHandler(handler)
    logger.info("New ansible-runner instance, logging initialized")



# cspell:ignore getpid, gmtime, msecs
"""Navigator entry point."""
from __future__ import annotations

import filecmp
import logging
import os
import signal
import sys

from copy import deepcopy
from curses import wrapper
from importlib.metadata import version
from importlib.util import find_spec
from pathlib import Path
from shutil import copyfile

from .action_defs import ActionReturn
from .action_defs import RunInteractiveReturn
from .action_defs import RunReturn
from .action_defs import RunStdoutReturn
from .action_runner import ActionRunner
from .actions import run_action_stdout
from .configuration_subsystem import Constants
from .configuration_subsystem import NavigatorConfiguration
from .configuration_subsystem.definitions import ApplicationConfiguration
from .image_manager import ImagePuller
from .initialization import error_and_exit_early
from .initialization import parse_and_update
from .logger import setup_logger
from .utils.compatibility import importlib_metadata
from .utils.definitions import ExitMessage
from .utils.definitions import ExitPrefix
from .utils.definitions import LogMessage
from .utils.functions import clear_screen
from .utils.functions import generate_cache_path
from .utils.packaged_data import path_to_file


__version__: Constants | str
try:
    from ._version import version as __version__
except ImportError:
    __version__ = Constants.NOT_SET


APP_NAME = "ansible-navigator"
PKG_NAME = "ansible_navigator"

logger = logging.getLogger(PKG_NAME)


def cache_scripts() -> None:
    """Cache the scripts used to introspect the container image."""
    scripts = ("catalog_collections.py", "image_introspect.py")
    for script in scripts:
        src = path_to_file(filename=script)
        cache_path = generate_cache_path(app_name=APP_NAME)
        cache_path.mkdir(parents=True, exist_ok=True)
        dst = cache_path / script
        message = f"No update required for {src} to {dst}"
        try:
            if not filecmp.cmp(src, dst):
                copyfile(src, dst)
                message = f"Updated {src} to {dst} (outdated)"
        except FileNotFoundError:
            copyfile(src, dst)
            message = f"Copied {src} to {dst} (missing)"
        logger.log(level=logging.DEBUG, msg=message)


def log_dependencies() -> list[LogMessage]:
    """Retrieve installed packages and log as debug.

    :returns: All packages, version and location
    """
    pkgs = []
    found = []
    messages: list[LogMessage] = []
    for _python_name, pkg_names in importlib_metadata.packages_distributions().items():
        for pkg_name in pkg_names:
            if pkg_name not in found:
                found.append(pkg_name)
                try:
                    spec = find_spec(pkg_name)
                except ModuleNotFoundError:
                    message = f"Package '{pkg_name}' is missing"
                    messages.append(LogMessage(level=logging.DEBUG, message=message))
                    continue
                _location = spec.origin if spec else ""
                _version = version(pkg_name)
                pkgs.append(f"{pkg_name}=={_version} {_location}")

    pkgs.sort()
    messages = [LogMessage(level=logging.DEBUG, message=pkg) for pkg in pkgs]
    return messages


def pull_image(args):
    """Pull the image if required.

    :param args: Copy of NavigatorConfiguration
    """
    image_puller = ImagePuller(
        container_engine=args.container_engine,
        image=args.execution_environment_image,
        arguments=args.pull_arguments,
        pull_policy=args.pull_policy,
    )
    image_puller.assess()
    if image_puller.assessment.exit_messages:
        error_and_exit_early(image_puller.assessment.exit_messages)
    if image_puller.assessment.pull_required:
        image_puller.prologue_stdout()
        image_puller.pull_stdout()


def run(args: ApplicationConfiguration) -> ActionReturn:
    """Run the appropriate subcommand.

    :param args: The current application settings
    :returns: A message to display and a return code
    """
    if args.mode == "stdout":
        try:
            result = run_action_stdout(args.app.replace("-", "_"), args)
            return result
        except KeyboardInterrupt:
            logger.warning("Dirty exit, killing the pid")
            os.kill(os.getpid(), signal.SIGTERM)
            return RunStdoutReturn(message="", return_code=1)
    elif args.mode == "interactive":
        try:
            clear_screen()
            wrapper(ActionRunner(args=args).run)
            return RunInteractiveReturn(message="", return_code=0)
        except KeyboardInterrupt:
            logger.warning("Dirty exit, killing the pid")
            os.kill(os.getpid(), signal.SIGTERM)
            return RunInteractiveReturn(message="", return_code=1)
    return RunReturn(message="", return_code=0)


def main():
    """Start application here."""
    messages: list[LogMessage] = log_dependencies()
    exit_messages: list[ExitMessage] = []

    args = deepcopy(NavigatorConfiguration)
    args.application_version = __version__
    args.internals.initializing = True
    messages.extend(args.internals.initialization_messages)
    exit_messages.extend(args.internals.initialization_exit_messages)

    # may have exit messages e.g., share directory
    # from instantiation of NavigatorConfiguration
    if not exit_messages:
        new_messages, new_exit_messages = parse_and_update(sys.argv[1:], args=args)
        messages.extend(new_messages)
        exit_messages.extend(new_exit_messages)

    # In case of errors, the configuration will have rolled back
    # but a viable log file is still needed, set to default since
    # it cannot be determined if the error is log file location related
    if exit_messages:
        args.entry("log_file").value.current = args.entry("log_file").value.default
        args.entry("log_level").value.current = "debug"
        exit_msg = f"Configuration failed, using default log file location. ({args.log_file})"
        exit_msg += f" Log level set to {args.log_level}"
        exit_messages.append(ExitMessage(message=exit_msg, prefix=ExitPrefix.NOTE))
        exit_msg = "Review the hints and log file to see what went wrong."
        exit_messages.append(ExitMessage(message=exit_msg, prefix=ExitPrefix.HINT))

    try:
        Path(args.log_file).touch()
        setup_logger(args)
    except Exception as exc:  # noqa: BLE001
        exit_msg = "The log file path or logging engine could not be setup."
        exit_msg += " No log file will be available, please check the log file"
        exit_msg += f" path setting. The error was {exc!s}"
        exit_messages.append(ExitMessage(message=exit_msg))
        error_and_exit_early(exit_messages=exit_messages)

    for entry in messages:
        logger.log(level=entry.level, msg=entry.message)

    if exit_messages:
        for exit_msg in exit_messages:
            logger.log(level=exit_msg.level, msg=exit_msg.message)
        error_and_exit_early(exit_messages=exit_messages)

    os.environ.setdefault("ESCDELAY", "25")

    if args.execution_environment:
        pull_image(args)
        cache_scripts()

    run_return = run(args)
    run_message = f"{run_return.message}\n"
    if run_return.return_code != 0 and run_return.message:
        sys.stderr.write(run_message)
        sys.exit(run_return.return_code)
    elif run_return.return_code != 0:
        sys.exit(run_return.return_code)
    elif run_return.message:
        sys.stdout.write(run_message)


if __name__ == "__main__":
    main()



"""The ability to perform migrations."""

from __future__ import annotations

from pathlib import Path

from ansible_navigator.utils.ansi import COLOR
from ansible_navigator.utils.ansi import IS_TTY
from ansible_navigator.utils.ansi import blank_line
from ansible_navigator.utils.ansi import info
from ansible_navigator.utils.ansi import prompt_enter
from ansible_navigator.utils.ansi import prompt_yn
from ansible_navigator.utils.ansi import success
from ansible_navigator.utils.ansi import warning

from .definitions import MigrationType
from .definitions import migrations


# isort: off
# pylint: disable=unused-import
# Migrations will run in the order they are imported
from .v1_v2_settings_file import V1V2SettingsFile

# isort: on


def run_all_migrations(migration_types: tuple[MigrationType], settings_file_str: str = "") -> None:
    """Run all migrations.

    :param migration_types: Type of migration
    :param settings_file_str: Path to the settings file
    """
    for migration_type in migration_types:
        run_migrations(settings_file_str, migration_type)


def run_migrations(settings_file_str: str, migration_type: MigrationType) -> None:
    """Run migrations of a specific type.

    :param settings_file_str: Path to the settings file
    :param migration_type: Type of migration
    """
    migrations_to_run = [
        migration()
        for migration in migrations
        if getattr(migration, "migration_type", None) == migration_type
    ]

    for migration in migrations_to_run:
        migration.check = True
        if getattr(migration, "migration_type", None) == MigrationType.SETTINGS_FILE:
            migration.settings_file_path = Path(settings_file_str)
        migration.run()

    migratable = [migration.name for migration in migrations_to_run if migration.needed_now]

    if migratable:
        tense = "are" if IS_TTY else "were"

        blank_line()
        warning(color=COLOR, message=f"The following version migrations {tense} required:")
        for migration in migrations_to_run:
            if migration.needed_now:
                info(color=COLOR, message=f"  - {migration.name}")

        if not IS_TTY:
            blank_line()
            warning(color=COLOR, message="Migration not possible without a TTY")
            return

        if prompt_yn(message="Do you want to run them all?"):
            for migration in migrations_to_run:
                blank_line()
                info(color=COLOR, message=f"{migration.name}:")
                migration.check = False
                migration.run()
            if not any(migration.needed_now for migration in migrations_to_run):
                blank_line()
                success(color=COLOR, message="Migration complete")
            else:
                blank_line()
                warning(
                    color=COLOR,
                    message="The following version migrations could not be completed:",
                )
                for migration in migrations_to_run:
                    if migration.needed_now:
                        info(color=COLOR, message=f" - {migration.name}")
        else:
            blank_line()
            warning(color=COLOR, message="Migration cancelled")
        blank_line()
        prompt_enter()



"""Introspect an execution environment image."""
from __future__ import annotations

import json
import os
import re
import subprocess
import sys
import threading

from queue import Queue
from types import SimpleNamespace
from typing import Any
from typing import Callable
from typing import Union


JSONTypes = Union[bool, int, str, dict, list]


class Command(SimpleNamespace):
    """Abstraction for a details about a shell command."""

    id_: str
    command: str
    parse: Callable
    stdout: str = ""
    stderr: str = ""
    details: list | dict | str = ""
    errors: list = []


def run_command(command: Command) -> None:
    """Run a command using subprocess.

    :param command: Details of the command to run
    """
    try:
        proc_out = subprocess.run(
            command.command,
            capture_output=True,
            check=True,
            text=True,
            shell=True,
        )
        command.stdout = proc_out.stdout
    except subprocess.CalledProcessError as exc:
        command.stderr = str(exc.stderr)
        command.errors = [str(exc.stderr)]


def worker(pending_queue: Queue, completed_queue: Queue) -> None:
    """Run a command from pending, parse, and place in completed.

    :param pending_queue: A queue with plugins to process
    :param completed_queue: The queue in which extracted documentation will be placed
    """
    while True:
        command = pending_queue.get()
        if command is None:
            break
        run_command(command)
        try:
            command.parse(command)
        except Exception as exc:  # noqa: BLE001
            command.errors = command.errors + [str(exc)]
        completed_queue.put(command)


class CommandRunner:
    """A command runner.

    Run commands using single or multiple processes.
    """

    def __init__(self):
        """Initialize the command runner."""
        self._completed_queue: Queue | None = None
        self._pending_queue: Queue | None = None

    def run_multi_thread(self, command_classes):
        """Run commands with multiple threads.

        Workers are started to read from pending queue.
        Exit when the number of results is equal to the number
        of commands needing to be run.

        :param command_classes: All command classes to be run
        :returns: The results from running all commands
        """
        if self._completed_queue is None:
            self._completed_queue = Queue()
        if self._pending_queue is None:
            self._pending_queue = Queue()
        results = {}
        all_commands = tuple(
            command for command_class in command_classes for command in command_class.commands
        )
        self.start_workers(all_commands)
        results = []
        while len(results) != len(all_commands):
            results.append(self._completed_queue.get())
        return results

    def start_workers(self, jobs):
        """Start workers and submit jobs to pending queue.

        :param jobs: The jobs to be run
        """
        worker_count = len(jobs)
        processes = []
        for _proc in range(worker_count):
            proc = threading.Thread(
                target=worker,
                args=(self._pending_queue, self._completed_queue),
            )
            processes.append(proc)
            proc.start()
        for job in jobs:
            self._pending_queue.put(job)
        for _proc in range(worker_count):
            self._pending_queue.put(None)
        for proc in processes:
            proc.join()


class CmdParser:
    """A base class for command parsers with common parsing functions."""

    @staticmethod
    def _strip(value: str) -> str:
        """Remove quotes, leading and trailing whitespace.

        :param value: The string to act on
        :returns: The string after removing quotes, leading and trailing whitespace
        """
        return value.strip('"').strip("'").strip()

    @staticmethod
    def re_partition(content, separator):
        """Partition a string using a regular expression.

        :param content: The content to partition
        :param separator: The separator to use for the partitioning
        :returns: The first partition, separator, and final partition
        """
        separator_match = re.search(separator, content)
        if not separator_match or content.startswith(" "):
            return "", "", content
        delim = separator_match.group(0)
        key, content = re.split(delim, content, maxsplit=1)
        return key, delim, content

    def splitter(self, lines, line_split, section_delim=None):
        """Split lines given a delimiter.

        :param lines: The lines to split
        :param line_split: The delimiter use for splitting each line
        :param section_delim: The separator between different packages
        :returns: All lines split on the delimiter
        """
        results = []
        result = {}
        current_key = ""
        while lines:
            line = lines.pop(0)
            key, delim, content = self.re_partition(line, line_split)
            content = self._strip(content)
            if section_delim and line == section_delim:
                results.append(result)
                result = {}
                continue
            if not delim and current_key:
                result[current_key] += f" {content}"
                continue
            current_key = key.lower().replace("_", "-").strip()
            # system_packages description field needs special handling
            if current_key == "description":
                description = []
                while lines:
                    line = lines.pop(0)
                    if line == section_delim:
                        break
                    description.append(line)
                if description:
                    result[current_key] = " ".join(description)
                else:
                    result[current_key] = "No description available"
                results.append(result)
                return result
            result[current_key] = content

        if result:
            results.append(result)
        return results


class AnsibleCollections(CmdParser):
    """Available ansible collections collector."""

    @property
    def commands(self):
        """Define the ansible-galaxy command to list ansible collections.

        :returns: The defined command
        """
        command = "ansible-galaxy collection list"
        return [
            Command(
                id_="ansible_collections",
                command=command,
                parse=self.parse,
            ),
        ]

    @staticmethod
    def parse(command: Command):
        """Parse the output of the ansible-galaxy command.

        :param command: The result of running the command
        """
        if "invalid choice: 'list'" in command.stderr:
            command.details = "This command is not supported with ansible 2.9."
            return
        collections = {}
        for line in command.stdout.splitlines():
            parts = line.split()
            if len(parts) == 2 and parts[1][0].isdigit():
                collections[parts[0].strip()] = parts[1].strip()
        command.details = collections


class AnsibleVersion(CmdParser):
    """Ansible version collector."""

    @property
    def commands(self) -> list[Command]:
        """Define the ansible command to get the version.

        :returns: The defined command
        """
        return [Command(id_="ansible_version", command="ansible --version", parse=self.parse)]

    @staticmethod
    def parse(command: Command) -> None:
        """Parse the output of the ansible command.

        :param command: The result of running the command
        """
        version = command.stdout.splitlines()[0]
        command.details = version


class OsRelease(CmdParser):
    """OS release information collector."""

    @property
    def commands(self) -> list[Command]:
        """Define the command to collect os release information.

        :returns: The defined command
        """
        return [Command(id_="os_release", command="cat /etc/os-release", parse=self.parse)]

    def parse(self, command) -> None:
        """Parse the output of the cat command.

        :param command: The result of running the command
        """
        parsed = self.splitter(command.stdout.splitlines(), "=")
        command.details = parsed


class PythonPackages(CmdParser):
    """Python package collector."""

    @property
    def commands(self) -> list[Command]:
        """Define the pip command to list installed pip packages.

        :returns: The defined command
        """
        pre = Command(
            id_="pip_freeze", command="/usr/bin/python3 -m pip freeze", parse=self.parse_freeze
        )
        run_command(pre)
        pre.parse(pre)
        pkgs = " ".join(pkg for pkg in pre.details[0]) if pre.details else ""
        return [
            Command(
                id_="python_packages",
                command=f"/usr/bin/python3 -m pip show {pkgs}",
                parse=self.parse,
            ),
        ]

    def parse(self, command):
        """Parse the output of the pip command.

        :param command: The result of running the command
        """
        parsed = self.splitter(command.stdout.splitlines(), line_split=":", section_delim="---")
        for pkg in parsed:
            for entry in ["required-by", "requires"]:
                if pkg[entry]:
                    pkg[entry] = [p.strip() for p in pkg[entry].split(",")]
                else:
                    pkg[entry] = []
        command.details = parsed

    def parse_freeze(self, command):
        """Parse the output of the pip freeze command, skipping editables.

        :param command: The result of running the command
        """
        lines = [line for line in command.stdout.splitlines() if not line.startswith("-e")]
        parsed = self.splitter(lines, "(==|@)")
        command.details = parsed


class RedhatRelease(CmdParser):
    """Red Hat release collector."""

    @property
    def commands(self) -> list[Command]:
        """Define the command to get the redhat release information.

        :returns: The defined command
        """
        return [Command(id_="redhat_release", command="cat /etc/redhat-release", parse=self.parse)]

    @staticmethod
    def parse(command):
        """Parse the output of the cat redhat release command.

        :param command: The result of running the command
        """
        parsed = command.stdout
        command.details = parsed


class SystemPackages(CmdParser):
    """System packages collector."""

    @property
    def commands(self) -> list[Command]:
        """Define the command to list system packages.

        :returns: The defined command
        """
        return [Command(id_="system_packages", command="rpm -qai", parse=self.parse)]

    def parse(self, command):
        """Parse the output of the rpm command.

        :param command: The result of running the command
        """
        packages = []
        package = []
        for line in command.stdout.splitlines():
            if re.match(r"^Name\s{2,}:", line) and package:
                packages.append(package)
                package = [line]
            else:
                package.append(line)
        if package:
            packages.append(package)

        parsed = []
        for package in packages:
            result = self.splitter(package, line_split=":")
            parsed.append(result)

        command.details = parsed


def main(serialize: bool = True) -> dict[str, JSONTypes] | None:
    """Enter the image introspection process.

    :param serialize: Whether to serialize the results
    :returns: The collected data or none if serialize is False
    """
    response: dict = {"errors": []}
    response["python_version"] = {"details": {"version": " ".join(sys.version.splitlines())}}
    response["environment_variables"] = {"details": dict(os.environ)}
    try:
        command_runner = CommandRunner()
        commands = [
            AnsibleCollections(),
            AnsibleVersion(),
            OsRelease(),
            RedhatRelease(),
            PythonPackages(),
            SystemPackages(),
        ]
        results = command_runner.run_multi_thread(commands)
        for result in results:
            result_as_dict = vars(result)
            result_as_dict.pop("parse")
            for key in list(result_as_dict.keys()):
                if key not in ["details", "errors"]:
                    result_as_dict[f"__{key}"] = result_as_dict[key]
                    result_as_dict.pop(key)
            response[result_as_dict["__id_"]] = result_as_dict
    except Exception as exc:  # noqa: BLE001
        response["errors"].append(str(exc))
    if serialize:
        print(json.dumps(response))
        return None
    return response


if __name__ == "__main__":
    main()



"""Collections subcommand implementation."""
from __future__ import annotations

import curses
import json
import os
import shlex
import sys

from copy import deepcopy
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Any
from typing import cast

from ansible_navigator.action_base import ActionBase
from ansible_navigator.action_defs import RunStdoutReturn
from ansible_navigator.app_public import AppPublic
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.content_defs import ContentFormat
from ansible_navigator.runner import Command
from ansible_navigator.steps import Step
from ansible_navigator.ui_framework import CursesLine
from ansible_navigator.ui_framework import CursesLinePart
from ansible_navigator.ui_framework import CursesLines
from ansible_navigator.ui_framework import Interaction
from ansible_navigator.ui_framework import nonblocking_notification
from ansible_navigator.ui_framework import warning_notification
from ansible_navigator.utils.functions import path_is_relative_to
from ansible_navigator.utils.functions import remove_dbl_un
from ansible_navigator.utils.key_value_store import KeyValueStore
from ansible_navigator.utils.print import print_to_stdout

from . import _actions as actions
from . import run_action


def color_menu(colno: int, colname: str, entry: dict[str, Any]) -> tuple[int, int]:
    """Provide a color for a collections menu entry in one column.

    :param colno: The column number
    :param colname: The column name
    :param entry: The menu entry
    :returns: The color and decoration
    """
    if entry.get("__shadowed") is True:
        return 8, 0
    if entry.get("__deprecated") == "True":
        return 9, 0
    return 2, 0


def content_heading(obj: Any, screen_w: int) -> CursesLines | None:
    """Create a heading for collection content.

    :param obj: The content going to be shown
    :param screen_w: The current screen width
    :returns: The heading
    """
    name = f"Image: {obj['full_name']}"
    description = f"Description: {obj['__description']}"
    padding = " " * (screen_w - len(description) + 1)
    line_1_part_1 = CursesLinePart(
        column=0,
        string=name,
        color=0,
        decoration=curses.A_NORMAL,
    )
    line_2_part_1 = CursesLinePart(
        column=0,
        string=description + padding,
        color=0,
        decoration=curses.A_UNDERLINE,
    )
    return CursesLines((CursesLine((line_1_part_1,)), CursesLine((line_2_part_1,))))


def filter_content_keys(obj: dict[Any, Any]) -> dict[Any, Any]:
    """Filter out some keys when showing collection content.

    :param obj: The object from which keys should be removed
    :returns: The object with keys removed
    """
    return {k: v for k, v in obj.items() if not k.startswith("__")}


@actions.register
class Action(ActionBase):
    """Collections subcommand implementation."""

    KEGEX = r"^collections(\s(?P<params>.*))?$"

    def __init__(self, args: ApplicationConfiguration):
        """Initialize the ``collections`` action.

        :param args: The current settings for the application
        """
        super().__init__(args=args, logger_name=__name__, name="collections")
        self._adjacent_collection_dir: str
        self._collection_cache: KeyValueStore
        self._collection_cache_path: str
        self._collection_scanned_paths: list = []
        self._collections: list = []
        self._stats: dict = {}

    def update(self) -> None:
        """Request calling app update, no collection update is required."""
        self._calling_app.update()

    def run(self, interaction: Interaction, app: AppPublic) -> Interaction | None:
        """Execute the ``collections`` request for mode interactive.

        :param interaction: The interaction from the user
        :param app: The app instance
        :returns: The pending :class:`~ansible_navigator.ui_framework.ui.Interaction` or
            :data:`None`
        """
        self._logger.debug("collections requested")
        self._prepare_to_run(app, interaction)
        self.stdout = self._calling_app.stdout

        notification = nonblocking_notification(
            messages=[
                "Collecting collection content, this may take a minute the first time...",
            ],
        )
        interaction.ui.show_form(notification)

        params = [self._name] + shlex.split(
            self._interaction.action.match.groupdict()["params"] or "",
        )

        args_updated = self._update_args(params=params, attach_cdc=True)
        if not args_updated:
            self._prepare_to_exit(interaction)
            return None

        if not isinstance(self._args.internals.collection_doc_cache, KeyValueStore):
            notification = warning_notification(
                messages=[
                    "Something has gone really wrong, the collection document cache is not",
                    "available.  This should not have happened. Please log an issue, and",
                    "include the contents of the log file.",
                ],
            )
            interaction.ui.show_form(notification)
            self._prepare_to_exit(interaction)
            return None

        self._collection_cache = self._args.internals.collection_doc_cache
        self._collection_cache_path = self._args.collection_doc_cache_path

        self._run_runner()

        if not self._collections:
            self._prepare_to_exit(interaction)
            self.notify_none()
            return None

        self.steps.append(self._build_main_menu())

        while True:
            self.update()
            self._take_step()

            if not self.steps:
                break

            if self.steps.current.name == "quit":
                return self.steps.current

        self._prepare_to_exit(interaction)
        return None

    def run_stdout(self) -> RunStdoutReturn:
        """Execute the ``collection`` request for mode stdout.

        :returns: The return code or 1. If the response from the runner invocation is None,
            indicates there is no console output to display, so assume an issue and return 1
            along with a message to review the logs.
        """
        self._logger.debug("collection requested in stdout mode")

        args_updated = self._update_args(params=[], attach_cdc=True)
        if not args_updated or not isinstance(
            self._args.internals.collection_doc_cache,
            KeyValueStore,
        ):
            msg = (
                "Failed to create collections cache, "
                "Please review the ansible-navigator log file for errors."
            )
            return RunStdoutReturn(message=msg, return_code=1)

        self._collection_cache = self._args.internals.collection_doc_cache
        self._collection_cache_path = self._args.collection_doc_cache_path

        self._run_runner()
        if not self._collections:
            msg = (
                "Failed to catalog collections, "
                "Please review the ansible-navigator log file for errors."
            )
            return RunStdoutReturn(message=msg, return_code=1)

        collections_info = self._parse_collection_info_stdout()

        print_to_stdout(
            content=collections_info,
            content_format=getattr(ContentFormat, self._args.format.upper()),
            use_color=self._args.display_color,
        )
        return RunStdoutReturn(message="", return_code=0)

    def notify_failed(self):
        """Notify collection cataloging failed."""
        msgs = ["humph. Something went really wrong while cataloging collections."]
        msgs.append("Details have been added to the log file")
        closing = ["[HINT] Please log an issue about this one, it shouldn't have happened"]
        warning = warning_notification(messages=msgs + closing)
        self._interaction.ui.show_form(warning)

    def notify_none(self):
        """Notify no collections were found."""
        msgs = ["humph. no collections were found in the following paths:"]
        paths = []
        for path in self._collection_scanned_paths:
            if path.startswith(self._args.internals.cache_path):
                continue
            if self._args.execution_environment:
                if path.startswith(self._adjacent_collection_dir):
                    paths.append(f"- {path} (bind_mount)")
                else:
                    paths.append(f"- {path} (contained)")
            else:
                paths.append(f"- {path}")
        closing = ["[HINT] Try installing some or try a different execution environment"]
        warning = warning_notification(messages=msgs + paths + closing)
        self._interaction.ui.show_form(warning)

    def _take_step(self) -> None:
        """Take a step based on the current step or step back."""
        result = None
        if isinstance(self.steps.current, Interaction):
            result = run_action(self.steps.current.name, self.app, self.steps.current)
        elif isinstance(self.steps.current, Step):
            if self.steps.current.show_func:
                current_index = self.steps.current.index
                self.steps.current.show_func()
                self.steps.current.index = current_index

            if self.steps.current.type == "menu":
                result = self._interaction.ui.show(
                    obj=self.steps.current.value,
                    columns=self.steps.current.columns,
                    color_menu_item=color_menu,
                )
            elif self.steps.current.type == "content":
                result = self._interaction.ui.show(
                    obj=self.steps.current.value,
                    index=self.steps.current.index,
                    content_heading=content_heading,
                    filter_content_keys=filter_content_keys,
                )

        if result is None:
            self.steps.back_one()
        else:
            self.steps.append(result)

    def _build_main_menu(self):
        """Build the menu of collections.

        :returns: The collections menu definition
        """
        if self._args.execution_environment:
            columns = ["__name", "__version", "__shadowed", "__type", "path"]
        else:
            columns = ["__name", "__version", "__shadowed", "path"]

        return Step(
            name="all_collections",
            columns=columns,
            select_func=self._build_collection_content_menu,
            step_type="menu",
            value=self._collections,
        )

    def _build_collection_content_menu(self):
        """Build the menu of plugins.

        :returns: The plugin menu definition
        """
        self._collection_cache.open_()
        selected_collection = self._collections[self.steps.current.index]
        collection_name = f"__{selected_collection['known_as']}"
        collection_contents = []
        for plugin_checksum, details in selected_collection["plugin_checksums"].items():
            try:
                plugin_json = self._collection_cache[plugin_checksum]
                loaded = json.loads(plugin_json)

                plugin = loaded["plugin"]
                if plugin["doc"] is not None:
                    if "name" in plugin["doc"]:
                        short_name = plugin["doc"]["name"]
                    else:
                        short_name = plugin["doc"][details["type"]]
                    plugin[collection_name] = short_name
                    plugin["full_name"] = f"{selected_collection['known_as']}.{short_name}"
                    plugin["__type"] = details["type"]
                    plugin["collection_info"] = selected_collection["collection_info"]
                    plugin["collection_info"]["name"] = selected_collection["known_as"]
                    plugin["collection_info"]["shadowed_by"] = selected_collection["hidden_by"]
                    plugin["collection_info"]["path"] = selected_collection["path"]

                    plugin["__added"] = plugin["doc"].get("version_added")
                    plugin["__description"] = plugin["doc"]["short_description"]

                    runtime_section = "modules" if details["type"] == "module" else details["type"]
                    plugin["__deprecated"] = "False"
                    try:
                        routing_info = selected_collection["runtime"]["plugin_routing"]
                        runtime_info = routing_info[runtime_section][short_name]
                        plugin["additional_information"] = runtime_info
                        if "deprecation" in runtime_info:
                            plugin["__deprecated"] = "True"
                    except KeyError:
                        plugin["additional_information"] = {}

                    collection_contents.append(plugin)
            except (KeyError, JSONDecodeError) as exc:
                self._logger.error("error loading plugin doc %s", details)
                self._logger.debug("error was %s", str(exc))

        self._collection_cache.close()

        for role in selected_collection["roles"]:
            role[collection_name] = role["short_name"]
            try:
                role["__description"] = role["info"]["galaxy_info"]["description"]
            except KeyError:
                role["__description"] = ""
            role["__deprecated"] = "Unknown"
            role["__added"] = "Unknown"
            role["__type"] = "role"
            collection_contents.append(role)

        collection_contents = sorted(collection_contents, key=lambda i: i[collection_name])

        return Step(
            name="all_collection_content",
            columns=[collection_name, "__type", "__added", "__deprecated", "__description"],
            select_func=self._build_collection_content,
            step_type="menu",
            value=collection_contents,
        )

    def _build_collection_content(self):
        """Build the content for one plugin.

        :returns: The plugin's content
        """
        return Step(
            name="collection_content",
            step_type="content",
            value=self.steps.current.value,
            index=self.steps.current.index,
        )

    def _run_runner(self) -> None:
        # pylint: disable=too-many-locals
        """Use the runner subsystem to catalog collections."""
        if isinstance(self._args.set_environment_variable, dict):
            set_environment_variable = deepcopy(self._args.set_environment_variable)
        else:
            set_environment_variable = {}
        set_environment_variable["ANSIBLE_NOCOLOR"] = "True"

        # We mount in the utils directory to /opt/ansible_navigator_utils
        # We do this so that we can access key_value_store (KVS) from within
        # the EE. If the Navigator user is overriding PYTHONPATH, we still need
        # to inject this utils directory into the PYTHONPATH. If not, we'll just
        # use the EE's default PYTHONPATH (if it exists) and just add our path
        # at the end.
        if self._args.execution_environment:
            ee_navigator_utils_mount = "/opt/ansible_navigator_utils"
            if "PYTHONPATH" in set_environment_variable:
                set_environment_variable["PYTHONPATH"].append(
                    f":{ee_navigator_utils_mount}",
                )
            else:
                set_environment_variable[
                    "PYTHONPATH"
                ] = f"${{PYTHONPATH}}:{ee_navigator_utils_mount}"
            self._logger.debug(
                "Execution Environment's PYTHONPATH is set to: %s",
                set_environment_variable["PYTHONPATH"],
            )

        kwargs = {
            "container_engine": self._args.container_engine,
            "execution_environment_image": self._args.execution_environment_image,
            "execution_environment": self._args.execution_environment,
            "navigator_mode": "interactive",
            "pass_environment_variable": self._args.pass_environment_variable,
            "set_environment_variable": set_environment_variable,
            "private_data_dir": self._args.ansible_runner_artifact_dir,
            "rotate_artifacts": self._args.ansible_runner_rotate_artifacts_count,
            "timeout": self._args.ansible_runner_timeout,
        }

        if isinstance(self._args.playbook, str):
            playbook_dir = os.path.dirname(self._args.playbook)
        else:
            playbook_dir = os.getcwd()

        if isinstance(self._args.execution_environment_volume_mounts, list):
            kwargs["container_volume_mounts"] = self._args.execution_environment_volume_mounts

        if isinstance(self._args.container_options, list):
            kwargs["container_options"] = self._args.container_options

        kwargs["host_cwd"] = playbook_dir

        self._adjacent_collection_dir = os.path.join(playbook_dir, "collections")
        cache_path = self._args.internals.cache_path

        pass_through_arg = [
            f"{cache_path}/catalog_collections.py",
            "-a",
            self._adjacent_collection_dir,
            "-c",
            self._collection_cache_path,
        ]

        kwargs["cmdline"] = pass_through_arg

        if self._args.execution_environment:
            self._logger.debug("running collections command with execution environment enabled")
            python_exec_path = "/usr/bin/python3"
            utils_lib = os.path.join(
                os.path.dirname(__file__),
                "..",
                "utils",
            )

            container_volume_mounts = [
                # cache directory which has introspection script
                f"{cache_path}:{cache_path}",
                # KVS library used by both Navigator and the introspection script
                f"{utils_lib}:/opt/ansible_navigator_utils",
            ]
            if os.path.exists(self._adjacent_collection_dir):
                container_volume_mounts.append(
                    f"{self._adjacent_collection_dir}:{self._adjacent_collection_dir}:z",
                )

            mount_doc_cache = True
            # Determine if the doc_cache is relative to the cache directory
            if path_is_relative_to(
                child=Path(self._args.collection_doc_cache_path), parent=(cache_path)
            ):
                mount_doc_cache = False

            # The playbook directory will be mounted as host_cwd, so don't duplicate
            if path_is_relative_to(
                child=Path(self._args.collection_doc_cache_path), parent=(Path(playbook_dir))
            ):
                mount_doc_cache = False

            if mount_doc_cache:
                container_volume_mounts.append(
                    f"{self._args.collection_doc_cache_path}:"
                    f"{self._args.collection_doc_cache_path}:z",
                )

            for volume_mount in container_volume_mounts:
                self._logger.debug("Adding volume mount to container invocation: %s", volume_mount)

            if "container_volume_mounts" in kwargs:
                kwargs["container_volume_mounts"] += container_volume_mounts
            else:
                kwargs["container_volume_mounts"] = container_volume_mounts

        else:
            self._logger.debug("running collections command locally")
            python_exec_path = sys.executable

        self._logger.debug(
            "Invoke runner with executable_cmd: %s and kwargs: %s",
            python_exec_path,
            kwargs,
        )
        _runner = Command(executable_cmd=python_exec_path, **kwargs)
        output, error, ret_code = _runner.run()

        if error:
            msg = f"Error while running catalog collection script: {error}"
            self._logger.error(msg)
        if ret_code:
            self._logger.error(output)
            self.notify_failed()
        if output:
            self._parse(output)

    def _parse(self, output) -> None:
        """Load and process the ``json`` output from the collection cataloging process.

        :param output: The output from the collection cataloging process
        :returns: Nothing
        """
        # pylint: disable=too-many-locals

        try:
            if not output.startswith("{"):
                _warnings, json_str = output.split("{", 1)
                json_str = "{" + json_str
            else:
                json_str = output
            parsed = json.loads(json_str)
            self._logger.debug("json loading output succeeded")
        except (JSONDecodeError, ValueError) as exc:
            self._logger.error("Unable to extract collection json from stdout")
            self._logger.debug("error json loading output: '%s'", str(exc))
            self._logger.debug(output)
            return None

        for error in parsed["errors"]:
            self._logger.error("%s %s", error["path"], error["error"])

        self._collections = sorted(
            list(parsed["collections"].values()),
            key=lambda i: i["known_as"],
        )
        volume_mounts = self.app.args.execution_environment_volume_mounts
        if isinstance(volume_mounts, list):
            tmp_list = []
            for mount in volume_mounts:
                source_str, destination_str = mount.split(":")[0:2]
                if not Path(source_str).is_dir():
                    continue
                dest_path = Path(destination_str)
                # /x/ansible_collections/co:/x/ansible_collections/co
                if dest_path.parent.name == "ansible_collections":
                    continue
                # /x/ansible_collections/co/ns:/x/ansible_collections/co/ns
                if dest_path.parent.parent.name == "ansible_collections":
                    continue
                # /x:/x
                if dest_path.name != "ansible_collections":
                    dest_path /= "ansible_collections"
                tmp_list.append(str(dest_path))
            dest_volume_mounts = tuple(tmp_list)
        else:
            dest_volume_mounts = tuple()

        for collection in self._collections:
            collection["__name"] = collection["known_as"]
            collection["__version"] = collection["collection_info"].get("version", "missing")
            collection["__shadowed"] = bool(collection["hidden_by"])
            if self._args.execution_environment:
                if collection["path"].startswith(dest_volume_mounts) or collection[
                    "path"
                ].startswith(self._adjacent_collection_dir):
                    collection["__type"] = "bind_mount"
                elif collection["path"].startswith(os.path.dirname(self._adjacent_collection_dir)):
                    collection["__type"] = "bind_mount"
                    error = (
                        f"{collection['known_as']} was mounted and catalogued in the"
                        " execution environment but was outside the adjacent 'collections'"
                        " directory. This may cause issues outside the local development"
                        " environment."
                    )
                    self._logger.error(error)
                else:
                    collection["__type"] = "contained"

        self._stats = parsed["stats"]

        if parsed.get("messages"):
            for msg in parsed["messages"]:
                self._logger.info("[catalog_collections]: %s", msg)

        self._logger.debug("catalog collections scan path: %s", parsed["collection_scan_paths"])
        self._collection_scanned_paths = parsed["collection_scan_paths"].split(":")
        for stat, value in self._stats.items():
            self._logger.debug("%s: %s", stat, value)

        if not parsed["collections"]:
            env = "execution" if self._args.execution_environment else "local"
            error = f"No collections found in {env} environment, searched in "
            error += parsed["collection_scan_paths"]
            self._logger.warning(error)

        return None

    def _get_collection_plugins_details(self, selected_collection: dict) -> dict:
        """Get plugin details for the given collection.

        :param selected_collection: The selected collection
        :returns: The plugin details like full-name, type and short description.
        """
        plugins_details: dict = {}

        for plugin_checksum, plugin_info in selected_collection["plugin_checksums"].items():
            plugin_type = plugin_info.get("type")
            if plugin_type not in plugins_details:
                plugins_details[plugin_type] = []

            plugin_json = self._collection_cache[plugin_checksum]
            loaded = json.loads(plugin_json)

            plugin = loaded.get("plugin")
            plugin_docs = {}
            plugin_path = os.path.join(
                selected_collection.get("path", ""),
                plugin_info.get("path", ""),
            )
            plugin_docs["path"] = plugin_path
            if plugin and plugin["doc"] is not None:
                try:
                    if "name" in plugin["doc"]:
                        short_name = plugin["doc"]["name"]
                    else:
                        short_name = plugin["doc"][plugin_type]
                except KeyError:
                    short_name = None
                if short_name is None:
                    plugin_docs["full_name"] = selected_collection["known_as"]
                else:
                    plugin_docs["full_name"] = f"{selected_collection['known_as']}.{short_name}"

                if "short_description" in plugin["doc"]:
                    plugin_docs["short_description"] = plugin["doc"]["short_description"]

            plugins_details[plugin_type].append(plugin_docs)

        return plugins_details

    def _parse_collection_info_stdout(self) -> dict:
        # pylint: disable=too-many-nested-blocks
        """Parse collection information from catalog collection cache.

        :returns: The collection information to be displayed on stdout
        """
        collections_info: dict = {
            "collections": [],
        }
        collection_exclude_keys = [
            "file_manifest_file",
            "format",
            "meta_source",
            "plugin_checksums",
            "runtime",
        ]
        roles_exclude_keys = ["readme"]
        self._collection_cache = cast(KeyValueStore, self._collection_cache)

        self._collection_cache.open_()
        for collection in self._collections:
            plugins_details = self._get_collection_plugins_details(collection)

            collection_stdout: dict = {}
            for info_name, info_value in collection.items():
                info_name = remove_dbl_un(info_name)
                if info_name in collection_exclude_keys:
                    continue

                if info_name == "roles":
                    collection_stdout["roles"] = []
                    for role in info_value:
                        updated_role_info = {}
                        for role_info_key, role_info_value in role.items():
                            if role_info_key in roles_exclude_keys:
                                continue
                            updated_role_info[role_info_key] = role_info_value
                        collection_stdout["roles"].append(updated_role_info)
                else:
                    collection_stdout[info_name] = info_value

            collection_stdout["plugins"] = plugins_details
            collections_info["collections"].append(collection_stdout)
        self._collection_cache.close()

        return collections_info



"""Template command implementation.

Processor of a template request at the single line prompt. e.g. {{ }}
"""
from __future__ import annotations

import html

from collections.abc import Mapping

from ansible_navigator.action_base import ActionBase
from ansible_navigator.app_public import AppPublic
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.content_defs import ContentFormat
from ansible_navigator.ui_framework import Interaction
from ansible_navigator.ui_framework import warning_notification
from ansible_navigator.utils.functions import remove_dbl_un
from ansible_navigator.utils.functions import templar

from . import _actions as actions


@actions.register
class Action(ActionBase):
    """Template command implementation."""

    KEGEX = r"^{{\s*(?P<params>.*?)\s*}}$"

    def __init__(self, args: ApplicationConfiguration):
        """Initialize the template action.

        :param args: The current settings for the application
        """
        super().__init__(args=args, logger_name=__name__, name="template")

    def run(self, interaction: Interaction, app: AppPublic) -> Interaction | None:
        """Execute the templating request for mode interactive.

        :param interaction: The interaction from the user
        :param app: The app instance
        :returns: The pending :class:`~ansible_navigator.ui_framework.ui.Interaction` or
            :data:`None`
        """
        self._logger.debug("template requested '%s'", interaction.action.value)
        self._prepare_to_run(app, interaction)

        type_msgs = ["Current content passed for templating is not a dictionary."]
        type_msgs.append("[HINT] Use 'this' to reference it (e.g. {{ this[0] }}")

        if interaction.content:
            if isinstance(interaction.content.showing, Mapping):
                template_vars = interaction.content.showing
                type_msgs = []
            else:
                template_vars = {"this": interaction.content.showing}

        elif interaction.menu:
            obj = [
                {remove_dbl_un(k): v for k, v in c.items() if k in interaction.menu.columns}
                for c in interaction.menu.current
            ]
            if interaction.ui.menu_filter():
                obj = [
                    e
                    for e in obj
                    if interaction.ui.menu_filter().search(" ".join(str(v) for v in e.values()))
                ]

            template_vars = {"this": obj}
        else:
            self._logger.error("No menu or content found")
            return None

        previous_scroll = interaction.ui.scroll()
        interaction.ui.scroll(0)

        errors, templated = templar(
            string=str(interaction.action.value),
            template_vars=template_vars,
        )
        if errors:
            msgs = ["Errors encountered while templating input"] + errors
            msgs.extend(type_msgs)
            interaction.ui.show_form(warning_notification(msgs))
            return None

        if isinstance(templated, str):
            templated = html.unescape(templated)

        requested = self._interaction.action.match.groupdict()["params"]
        if requested == "examples":
            content_format = ContentFormat.YAML_TXT
        elif requested == "readme":
            content_format = ContentFormat.MARKDOWN
        elif isinstance(templated, str):
            content_format = ContentFormat.TXT
        else:
            content_format = None

        while True:
            app.update()
            next_interaction: Interaction = interaction.ui.show(
                obj=templated,
                content_format=content_format,
            )
            if next_interaction.name != "refresh":
                break

        interaction.ui.scroll(previous_scroll)
        self._prepare_to_exit(interaction)
        return next_interaction
